# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

import ccxt.async_support
from ccxt.async_support.base.ws.cache import ArrayCache, ArrayCacheBySymbolById, ArrayCacheByTimestamp
import hashlib
from ccxt.base.types import Balances, Int, Order, OrderBook, Str, Ticker, Trade
from ccxt.async_support.base.ws.client import Client
from typing import List
from ccxt.base.errors import AuthenticationError
from ccxt.base.precise import Precise


class phemex(ccxt.async_support.phemex):

    def describe(self):
        return self.deep_extend(super(phemex, self).describe(), {
            'has': {
                'ws': True,
                'watchTicker': True,
                'watchTickers': False,  # for now
                'watchTrades': True,
                'watchMyTrades': True,
                'watchOrders': True,
                'watchOrderBook': True,
                'watchOHLCV': True,
                'watchPositions': None,  # TODO
                # mutli-endpoints are not supported: https://github.com/ccxt/ccxt/pull/21490
                'watchOrderBookForSymbols': False,
                'watchTradesForSymbols': False,
                'watchOHLCVForSymbols': False,
            },
            'urls': {
                'test': {
                    'ws': 'wss://testnet-api.phemex.com/ws',
                },
                'api': {
                    'ws': 'wss://ws.phemex.com',
                },
            },
            'options': {
                'tradesLimit': 1000,
                'OHLCVLimit': 1000,
            },
            'streaming': {
                'keepAlive': 10000,
            },
        })

    def from_en(self, en, scale):
        if en is None:
            return None
        precise = Precise(en)
        precise.decimals = self.sum(precise.decimals, scale)
        precise.reduce()
        return str(precise)

    def from_ep(self, ep, market=None):
        if (ep is None) or (market is None):
            return ep
        return self.from_en(ep, self.safe_integer(market, 'priceScale'))

    def from_ev(self, ev, market=None):
        if (ev is None) or (market is None):
            return ev
        return self.from_en(ev, self.safe_integer(market, 'valueScale'))

    def from_er(self, er, market=None):
        if (er is None) or (market is None):
            return er
        return self.from_en(er, self.safe_integer(market, 'ratioScale'))

    def request_id(self):
        requestId = self.sum(self.safe_integer(self.options, 'requestId', 0), 1)
        self.options['requestId'] = requestId
        return requestId

    def parse_swap_ticker(self, ticker, market=None):
        #
        #     {
        #         "close": 442800,
        #         "fundingRate": 10000,
        #         "high": 445400,
        #         "indexPrice": 442621,
        #         "low": 428400,
        #         "markPrice": 442659,
        #         "open": 432200,
        #         "openInterest": 744183,
        #         "predFundingRate": 10000,
        #         "symbol": "LTCUSD",
        #         "turnover": 8133238294,
        #         "volume": 934292
        #     }
        #
        marketId = self.safe_string(ticker, 'symbol')
        market = self.safe_market(marketId, market)
        symbol = market['symbol']
        timestamp = self.safe_integer_product(ticker, 'timestamp', 0.000001)
        lastString = self.from_ep(self.safe_string(ticker, 'close'), market)
        last = self.parse_number(lastString)
        quoteVolume = self.parse_number(self.from_ev(self.safe_string(ticker, 'turnover'), market))
        baseVolume = self.parse_number(self.from_ev(self.safe_string(ticker, 'volume'), market))
        change = None
        percentage = None
        average = None
        openString = self.omit_zero(self.from_ep(self.safe_string(ticker, 'open'), market))
        open = self.parse_number(openString)
        if (openString is not None) and (lastString is not None):
            change = self.parse_number(Precise.string_sub(lastString, openString))
            average = self.parse_number(Precise.string_div(Precise.string_add(lastString, openString), '2'))
            percentage = self.parse_number(Precise.string_mul(Precise.string_sub(Precise.string_div(lastString, openString), '1'), '100'))
        result: dict = {
            'symbol': symbol,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'high': self.parse_number(self.from_ep(self.safe_string(ticker, 'high'), market)),
            'low': self.parse_number(self.from_ep(self.safe_string(ticker, 'low'), market)),
            'bid': None,
            'bidVolume': None,
            'ask': None,
            'askVolume': None,
            'vwap': None,
            'open': open,
            'close': last,
            'last': last,
            'previousClose': None,  # previous day close
            'change': change,
            'percentage': percentage,
            'average': average,
            'baseVolume': baseVolume,
            'quoteVolume': quoteVolume,
            'info': ticker,
        }
        return result

    def parse_perpetual_ticker(self, ticker, market=None):
        #
        #    [
        #        "STXUSDT",
        #        "0.64649",
        #        "0.8628",
        #        "0.61215",
        #        "0.71737",
        #        "4519387",
        #        "3210827.98166",
        #        "697635",
        #        "0.71720205",
        #        "0.71720205",
        #        "0.0001",
        #        "0.0001",
        #    ]
        #
        marketId = self.safe_string(ticker, 0)
        market = self.safe_market(marketId, market)
        symbol = market['symbol']
        lastString = self.from_ep(self.safe_string(ticker, 4), market)
        last = self.parse_number(lastString)
        quoteVolume = self.parse_number(self.from_ev(self.safe_string(ticker, 6), market))
        baseVolume = self.parse_number(self.from_ev(self.safe_string(ticker, 5), market))
        change = None
        percentage = None
        average = None
        openString = self.omit_zero(self.from_ep(self.safe_string(ticker, 1), market))
        open = self.parse_number(openString)
        if (openString is not None) and (lastString is not None):
            change = self.parse_number(Precise.string_sub(lastString, openString))
            average = self.parse_number(Precise.string_div(Precise.string_add(lastString, openString), '2'))
            percentage = self.parse_number(Precise.string_mul(Precise.string_sub(Precise.string_div(lastString, openString), '1'), '100'))
        result: dict = {
            'symbol': symbol,
            'timestamp': None,
            'datetime': None,
            'high': self.parse_number(self.from_ep(self.safe_string(ticker, 2), market)),
            'low': self.parse_number(self.from_ep(self.safe_string(ticker, 3), market)),
            'bid': None,
            'bidVolume': None,
            'ask': None,
            'askVolume': None,
            'vwap': None,
            'open': open,
            'close': last,
            'last': last,
            'previousClose': None,  # previous day close
            'change': change,
            'percentage': percentage,
            'average': average,
            'baseVolume': baseVolume,
            'quoteVolume': quoteVolume,
            'info': ticker,
        }
        return result

    def handle_ticker(self, client: Client, message):
        #
        #     {
        #         "spot_market24h": {
        #             "askEp": 958148000000,
        #             "bidEp": 957884000000,
        #             "highEp": 962000000000,
        #             "lastEp": 958220000000,
        #             "lowEp": 928049000000,
        #             "openEp": 935597000000,
        #             "symbol": "sBTCUSDT",
        #             "turnoverEv": 146074214388978,
        #             "volumeEv": 15492228900
        #         },
        #         "timestamp": 1592847265888272100
        #     }
        #
        # swap
        #
        #     {
        #         "market24h": {
        #             "close": 442800,
        #             "fundingRate": 10000,
        #             "high": 445400,
        #             "indexPrice": 442621,
        #             "low": 428400,
        #             "markPrice": 442659,
        #             "open": 432200,
        #             "openInterest": 744183,
        #             "predFundingRate": 10000,
        #             "symbol": "LTCUSD",
        #             "turnover": 8133238294,
        #             "volume": 934292
        #         },
        #         "timestamp": 1592845585373374500
        #     }
        #
        # perpetual
        #
        #    {
        #        "data": [
        #            [
        #                "STXUSDT",
        #                "0.64649",
        #                "0.8628",
        #                "0.61215",
        #                "0.71737",
        #                "4519387",
        #                "3210827.98166",
        #                "697635",
        #                "0.71720205",
        #                "0.71720205",
        #                "0.0001",
        #                "0.0001",
        #            ],
        #            ...
        #        ],
        #        "fields": [
        #            "symbol",
        #            "openRp",
        #            "highRp",
        #            "lowRp",
        #            "lastRp",
        #            "volumeRq",
        #            "turnoverRv",
        #            "openInterestRv",
        #            "indexRp",
        #            "markRp",
        #            "fundingRateRr",
        #            "predFundingRateRr",
        #        ],
        #        "method": "perp_market24h_pack_p.update",
        #        "timestamp": "1677094918686806209",
        #        "type": "snapshot",
        #    }
        #
        tickers = []
        if 'market24h' in message:
            ticker = self.safe_value(message, 'market24h')
            tickers.append(self.parse_swap_ticker(ticker))
        elif 'spot_market24h' in message:
            ticker = self.safe_value(message, 'spot_market24h')
            tickers.append(self.parse_ticker(ticker))
        elif 'data' in message:
            data = self.safe_value(message, 'data', [])
            for i in range(0, len(data)):
                tickers.append(self.parse_perpetual_ticker(data[i]))
        for i in range(0, len(tickers)):
            ticker = tickers[i]
            symbol = ticker['symbol']
            messageHash = 'ticker:' + symbol
            timestamp = self.safe_integer_product(message, 'timestamp', 0.000001)
            ticker['timestamp'] = timestamp
            ticker['datetime'] = self.iso8601(timestamp)
            self.tickers[symbol] = ticker
            client.resolve(ticker, messageHash)

    async def watch_balance(self, params={}) -> Balances:
        """
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Hedged-Perpetual-API.md#subscribe-account-order-position-aop
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#subscribe-account-order-position-aop
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Spot-API-en.md#subscribe-wallet-order-messages
        watch balance and get the amount of funds available for trading or funds locked in orders
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.settle]: set to USDT to use hedged perpetual api
        :returns dict: a `balance structure <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        await self.load_markets()
        type = None
        type, params = self.handle_market_type_and_params('watchBalance', None, params)
        usePerpetualApi = self.safe_string(params, 'settle') == 'USDT'
        messageHash = ':balance'
        messageHash = 'perpetual' + messageHash if usePerpetualApi else type + messageHash
        return await self.subscribe_private(type, messageHash, params)

    def handle_balance(self, type, client, message):
        # spot
        #    [
        #       {
        #           "balanceEv": 0,
        #           "currency": "BTC",
        #           "lastUpdateTimeNs": "1650442638722099092",
        #           "lockedTradingBalanceEv": 0,
        #           "lockedWithdrawEv": 0,
        #           "userID": 2647224
        #         },
        #         {
        #           "balanceEv": 1154232337,
        #           "currency": "USDT",
        #           "lastUpdateTimeNs": "1650442617610017597",
        #           "lockedTradingBalanceEv": 0,
        #           "lockedWithdrawEv": 0,
        #           "userID": 2647224
        #         }
        #    ]
        # swap
        #    [
        #        {
        #            "accountBalanceEv": 0,
        #            "accountID": 26472240001,
        #            "bonusBalanceEv": 0,
        #            "currency": "BTC",
        #            "totalUsedBalanceEv": 0,
        #            "userID": 2647224
        #        }
        #    ]
        # perpetual
        #    [
        #        {
        #            "accountBalanceRv": "1508.452588802237",
        #            "accountID": 9328670003,
        #            "bonusBalanceRv": "0",
        #            "currency": "USDT",
        #            "totalUsedBalanceRv": "343.132599666883",
        #            "userID": 932867
        #        }
        #    ]
        #
        self.balance['info'] = message
        for i in range(0, len(message)):
            balance = message[i]
            currencyId = self.safe_string(balance, 'currency')
            code = self.safe_currency_code(currencyId)
            currency = self.safe_value(self.currencies, code, {})
            scale = self.safe_integer(currency, 'valueScale', 8)
            account = self.account()
            used = self.safe_string(balance, 'totalUsedBalanceRv')
            if used is None:
                usedEv = self.safe_string(balance, 'totalUsedBalanceEv')
                if usedEv is None:
                    lockedTradingBalanceEv = self.safe_string(balance, 'lockedTradingBalanceEv')
                    lockedWithdrawEv = self.safe_string_2(balance, 'lockedWithdrawEv', 'lockedWithdrawRv')
                    usedEv = Precise.string_add(lockedTradingBalanceEv, lockedWithdrawEv)
                used = self.from_en(usedEv, scale)
            total = self.safe_string(balance, 'accountBalanceRv')
            if total is None:
                totalEv = self.safe_string_2(balance, 'accountBalanceEv', 'balanceEv')
                total = self.from_en(totalEv, scale)
            account['used'] = used
            account['total'] = total
            self.balance[code] = account
            self.balance = self.safe_balance(self.balance)
        messageHash = type + ':balance'
        client.resolve(self.balance, messageHash)

    def handle_trades(self, client: Client, message):
        #
        #     {
        #         "sequence": 1795484727,
        #         "symbol": "sBTCUSDT",
        #         "trades": [
        #             [1592891002064516600, "Buy", 964020000000, 1431000],
        #             [1592890978987934500, "Sell", 963704000000, 1401800],
        #             [1592890972918701800, "Buy", 963938000000, 2018600],
        #         ],
        #         "type": "snapshot"
        #     }
        #  perpetual
        #     {
        #         "sequence": 1230197759,
        #         "symbol": "BTCUSDT",
        #         "trades_p": [
        #             [
        #                 1677094244729433000,
        #                 "Buy",
        #                 "23800.4",
        #                 "2.455",
        #             ],
        #         ],
        #         "type": "snapshot",
        #     }
        #
        name = 'trade'
        marketId = self.safe_string(message, 'symbol')
        market = self.safe_market(marketId)
        symbol = market['symbol']
        messageHash = name + ':' + symbol
        stored = self.safe_value(self.trades, symbol)
        if stored is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            stored = ArrayCache(limit)
            self.trades[symbol] = stored
        trades = self.safe_value_2(message, 'trades', 'trades_p', [])
        parsed = self.parse_trades(trades, market)
        for i in range(0, len(parsed)):
            stored.append(parsed[i])
        client.resolve(stored, messageHash)

    def handle_ohlcv(self, client: Client, message):
        #
        #     {
        #         "kline": [
        #             [1592905200, 60, 960688000000, 960709000000, 960709000000, 960400000000, 960400000000, 848100, 8146756046],
        #             [1592905140, 60, 960718000000, 960716000000, 960717000000, 960560000000, 960688000000, 4284900, 41163743512],
        #             [1592905080, 60, 960513000000, 960684000000, 960718000000, 960684000000, 960718000000, 4880500, 46887494349],
        #         ],
        #         "sequence": 1804401474,
        #         "symbol": "sBTCUSDT",
        #         "type": "snapshot"
        #     }
        # perpetual
        #     {
        #         "kline_p": [
        #             [
        #                 1677094560,
        #                 60,
        #                 "23746.2",
        #                 "23746.1",
        #                 "23757.6",
        #                 "23736.9",
        #                 "23754.8",
        #                 "34.273",
        #                 "813910.208",
        #             ],
        #         ],
        #         "sequence": 1230786017,
        #         "symbol": "BTCUSDT",
        #         "type": "incremental",
        #     }
        #
        marketId = self.safe_string(message, 'symbol')
        market = self.safe_market(marketId)
        symbol = market['symbol']
        candles = self.safe_value_2(message, 'kline', 'kline_p', [])
        first = self.safe_value(candles, 0, [])
        interval = self.safe_string(first, 1)
        timeframe = self.find_timeframe(interval)
        if timeframe is not None:
            messageHash = 'kline:' + timeframe + ':' + symbol
            ohlcvs = self.parse_ohlcvs(candles, market)
            self.ohlcvs[symbol] = self.safe_value(self.ohlcvs, symbol, {})
            stored = self.safe_value(self.ohlcvs[symbol], timeframe)
            if stored is None:
                limit = self.safe_integer(self.options, 'OHLCVLimit', 1000)
                stored = ArrayCacheByTimestamp(limit)
                self.ohlcvs[symbol][timeframe] = stored
            for i in range(0, len(ohlcvs)):
                candle = ohlcvs[i]
                stored.append(candle)
            client.resolve(stored, messageHash)

    async def watch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Hedged-Perpetual-API.md#subscribe-24-hours-ticker
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#subscribe-24-hours-ticker
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Spot-API-en.md#subscribe-24-hours-ticker
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        isSwap = market['swap']
        settleIsUSDT = market['settle'] == 'USDT'
        name = 'spot_market24h'
        if isSwap:
            name = 'perp_market24h_pack_p' if settleIsUSDT else 'market24h'
        url = self.urls['api']['ws']
        requestId = self.request_id()
        subscriptionHash = name + '.subscribe'
        messageHash = 'ticker:' + symbol
        subscribe: dict = {
            'method': subscriptionHash,
            'id': requestId,
            'params': [],
        }
        request = self.deep_extend(subscribe, params)
        return await self.watch(url, messageHash, request, subscriptionHash)

    async def watch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Hedged-Perpetual-API.md#subscribe-trade
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#subscribe-trade
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Spot-API-en.md#subscribe-trade
        get the list of most recent trades for a particular symbol
        :param str symbol: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        url = self.urls['api']['ws']
        requestId = self.request_id()
        isSwap = market['swap']
        settleIsUSDT = market['settle'] == 'USDT'
        name = 'trade_p' if (isSwap and settleIsUSDT) else 'trade'
        messageHash = 'trade:' + symbol
        method = name + '.subscribe'
        subscribe: dict = {
            'method': method,
            'id': requestId,
            'params': [
                market['id'],
            ],
        }
        request = self.deep_extend(subscribe, params)
        trades = await self.watch(url, messageHash, request, messageHash)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp', True)

    async def watch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Spot-API-en.md#subscribe-orderbook
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Hedged-Perpetual-API.md#subscribe-orderbook-for-new-model
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#subscribe-30-levels-orderbook
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#subscribe-full-orderbook
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        url = self.urls['api']['ws']
        requestId = self.request_id()
        isSwap = market['swap']
        settleIsUSDT = market['settle'] == 'USDT'
        name = 'orderbook_p' if (isSwap and settleIsUSDT) else 'orderbook'
        messageHash = 'orderbook:' + symbol
        method = name + '.subscribe'
        subscribe: dict = {
            'method': method,
            'id': requestId,
            'params': [
                market['id'],
            ],
        }
        request = self.deep_extend(subscribe, params)
        orderbook = await self.watch(url, messageHash, request, messageHash)
        return orderbook.limit()

    async def watch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Hedged-Perpetual-API.md#subscribe-kline
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#subscribe-kline
        :see: https://github.com/phemex/phemex-api-docs/blob/master/Public-Spot-API-en.md#subscribe-kline
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        url = self.urls['api']['ws']
        requestId = self.request_id()
        isSwap = market['swap']
        settleIsUSDT = market['settle'] == 'USDT'
        name = 'kline_p' if (isSwap and settleIsUSDT) else 'kline'
        messageHash = 'kline:' + timeframe + ':' + symbol
        method = name + '.subscribe'
        subscribe: dict = {
            'method': method,
            'id': requestId,
            'params': [
                market['id'],
                self.safe_integer(self.timeframes, timeframe),
            ],
        }
        request = self.deep_extend(subscribe, params)
        ohlcv = await self.watch(url, messageHash, request, messageHash)
        if self.newUpdates:
            limit = ohlcv.getLimit(symbol, limit)
        return self.filter_by_since_limit(ohlcv, since, limit, 0, True)

    def custom_handle_delta(self, bookside, delta, market=None):
        bidAsk = self.custom_parse_bid_ask(delta, 0, 1, market)
        bookside.storeArray(bidAsk)

    def custom_handle_deltas(self, bookside, deltas, market=None):
        for i in range(0, len(deltas)):
            self.custom_handle_delta(bookside, deltas[i], market)

    def handle_order_book(self, client: Client, message):
        #
        #     {
        #         "book": {
        #             "asks": [
        #                 [960316000000, 6993800],
        #                 [960318000000, 13183000],
        #                 [960319000000, 9170200],
        #             ],
        #             "bids": [
        #                 [959941000000, 8385300],
        #                 [959939000000, 10296600],
        #                 [959930000000, 3672400],
        #             ]
        #         },
        #         "depth": 30,
        #         "sequence": 1805784701,
        #         "symbol": "sBTCUSDT",
        #         "timestamp": 1592908460404461600,
        #         "type": "snapshot"
        #     }
        #  perpetual
        #    {
        #        "depth": 30,
        #        "orderbook_p": {
        #            "asks": [
        #                [
        #                    "23788.5",
        #                    "0.13",
        #                ],
        #            ],
        #            "bids": [
        #                [
        #                    "23787.8",
        #                    "1.836",
        #                ],
        #            ],
        #        },
        #        "sequence": 1230347368,
        #        "symbol": "BTCUSDT",
        #        "timestamp": "1677093457306978852",
        #        "type": "snapshot",
        #    }
        #
        marketId = self.safe_string(message, 'symbol')
        market = self.safe_market(marketId)
        symbol = market['symbol']
        type = self.safe_string(message, 'type')
        depth = self.safe_integer(message, 'depth')
        name = 'orderbook'
        messageHash = name + ':' + symbol
        nonce = self.safe_integer(message, 'sequence')
        timestamp = self.safe_integer_product(message, 'timestamp', 0.000001)
        if type == 'snapshot':
            book = self.safe_value_2(message, 'book', 'orderbook_p', {})
            snapshot = self.custom_parse_order_book(book, symbol, timestamp, 'bids', 'asks', 0, 1, market)
            snapshot['nonce'] = nonce
            orderbook = self.order_book(snapshot, depth)
            self.orderbooks[symbol] = orderbook
            client.resolve(orderbook, messageHash)
        else:
            if symbol in self.orderbooks:
                orderbook = self.orderbooks[symbol]
                changes = self.safe_dict_2(message, 'book', 'orderbook_p', {})
                asks = self.safe_list(changes, 'asks', [])
                bids = self.safe_list(changes, 'bids', [])
                self.custom_handle_deltas(orderbook['asks'], asks, market)
                self.custom_handle_deltas(orderbook['bids'], bids, market)
                orderbook['nonce'] = nonce
                orderbook['timestamp'] = timestamp
                orderbook['datetime'] = self.iso8601(timestamp)
                self.orderbooks[symbol] = orderbook
                client.resolve(orderbook, messageHash)

    async def watch_my_trades(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        watches information on multiple trades made by the user
        :param str symbol: unified market symbol of the market trades were made in
        :param int [since]: the earliest time in ms to fetch trades for
        :param int [limit]: the maximum number of trade structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of [trade structures]{@link https://docs.ccxt.com/#/?id=trade-structure
        """
        await self.load_markets()
        market = None
        type = None
        messageHash = 'trades:'
        if symbol is not None:
            market = self.market(symbol)
            symbol = market['symbol']
            messageHash = messageHash + market['symbol']
            if market['settle'] == 'USDT':
                params = self.extend(params)
                params['settle'] = 'USDT'
        type, params = self.handle_market_type_and_params('watchMyTrades', market, params)
        if symbol is None:
            settle = self.safe_string(params, 'settle')
            messageHash = (messageHash + 'perpetual') if (settle == 'USDT') else (messageHash + type)
        trades = await self.subscribe_private(type, messageHash, params)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        return self.filter_by_symbol_since_limit(trades, symbol, since, limit, True)

    def handle_my_trades(self, client: Client, message):
        #
        # swap
        #    [
        #        {
        #            "avgPriceEp":4138763000000,
        #            "baseCurrency":"BTC",
        #            "baseQtyEv":0,
        #            "clOrdID":"7956e0be-e8be-93a0-2887-ca504d85cda2",
        #            "execBaseQtyEv":30100,
        #            "execFeeEv":31,
        #            "execID":"d3b10cfa-84e3-5752-828e-78a79617e598",
        #            "execPriceEp":4138763000000,
        #            "execQuoteQtyEv":1245767663,
        #            "feeCurrency":"BTC",
        #            "lastLiquidityInd":"RemovedLiquidity",
        #            "ordType":"Market",
        #            "orderID":"34a4b1a8-ac3a-4580-b3e6-a6d039f27195",
        #            "priceEp":4549022000000,
        #            "qtyType":"ByQuote",
        #            "quoteCurrency":"USDT",
        #            "quoteQtyEv":1248000000,
        #            "side":"Buy",
        #            "symbol":"sBTCUSDT",
        #            "tradeType":"Trade",
        #            "transactTimeNs":"1650442617609928764",
        #            "userID":2647224
        #        }
        #    ]
        # perpetual
        #    [
        #        {
        #            "accountID": 9328670003,
        #            "action": "New",
        #            "actionBy": "ByUser",
        #            "actionTimeNs": 1666858780876924611,
        #            "addedSeq": 77751555,
        #            "apRp": "0",
        #            "bonusChangedAmountRv": "0",
        #            "bpRp": "0",
        #            "clOrdID": "c0327a7d-9064-62a9-28f6-2db9aaaa04e0",
        #            "closedPnlRv": "0",
        #            "closedSize": "0",
        #            "code": 0,
        #            "cumFeeRv": "0",
        #            "cumQty": "0",
        #            "cumValueRv": "0",
        #            "curAccBalanceRv": "1508.489893982237",
        #            "curAssignedPosBalanceRv": "24.62786650928",
        #            "curBonusBalanceRv": "0",
        #            "curLeverageRr": "-10",
        #            "curPosSide": "Buy",
        #            "curPosSize": "0.043",
        #            "curPosTerm": 1,
        #            "curPosValueRv": "894.0689",
        #            "curRiskLimitRv": "1000000",
        #            "currency": "USDT",
        #            "cxlRejReason": 0,
        #            "displayQty": "0.003",
        #            "execFeeRv": "0",
        #            "execID": "00000000-0000-0000-0000-000000000000",
        #            "execPriceRp": "20723.7",
        #            "execQty": "0",
        #            "execSeq": 77751555,
        #            "execStatus": "New",
        #            "execValueRv": "0",
        #            "feeRateRr": "0",
        #            "leavesQty": "0.003",
        #            "leavesValueRv": "63.4503",
        #            "message": "No error",
        #            "ordStatus": "New",
        #            "ordType": "Market",
        #            "orderID": "fa64c6f2-47a4-4929-aab4-b7fa9bbc4323",
        #            "orderQty": "0.003",
        #            "pegOffsetValueRp": "0",
        #            "posSide": "Long",
        #            "priceRp": "21150.1",
        #            "relatedPosTerm": 1,
        #            "relatedReqNum": 11,
        #            "side": "Buy",
        #            "slTrigger": "ByMarkPrice",
        #            "stopLossRp": "0",
        #            "stopPxRp": "0",
        #            "symbol": "BTCUSDT",
        #            "takeProfitRp": "0",
        #            "timeInForce": "ImmediateOrCancel",
        #            "tpTrigger": "ByLastPrice",
        #            "tradeType": "Amend",
        #            "transactTimeNs": 1666858780881545305,
        #            "userID": 932867
        #        },
        #        ...
        #    ]
        #
        channel = 'trades'
        tradesLength = len(message)
        if tradesLength == 0:
            return
        cachedTrades = self.myTrades
        if cachedTrades is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            cachedTrades = ArrayCacheBySymbolById(limit)
        marketIds: dict = {}
        type = None
        for i in range(0, len(message)):
            rawTrade = message[i]
            marketId = self.safe_string(rawTrade, 'symbol')
            market = self.safe_market(marketId)
            parsed = self.parse_trade(rawTrade)
            cachedTrades.append(parsed)
            symbol = parsed['symbol']
            if type is None:
                type = 'perpetual' if (market['settle'] == 'USDT') else market['type']
            marketIds[symbol] = True
        keys = list(marketIds.keys())
        for i in range(0, len(keys)):
            market = keys[i]
            hash = channel + ':' + market
            client.resolve(cachedTrades, hash)
        # generic subscription
        messageHash = channel + ':' + type
        client.resolve(cachedTrades, messageHash)

    async def watch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        watches information on multiple orders made by the user
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        await self.load_markets()
        messageHash = 'orders:'
        market = None
        type = None
        if symbol is not None:
            market = self.market(symbol)
            symbol = market['symbol']
            messageHash = messageHash + market['symbol']
            if market['settle'] == 'USDT':
                params = self.extend(params)
                params['settle'] = 'USDT'
        type, params = self.handle_market_type_and_params('watchOrders', market, params)
        isUSDTSettled = self.safe_string(params, 'settle') == 'USDT'
        if symbol is None:
            messageHash = (messageHash + 'perpetual') if (isUSDTSettled) else (messageHash + type)
        orders = await self.subscribe_private(type, messageHash, params)
        if self.newUpdates:
            limit = orders.getLimit(symbol, limit)
        return self.filter_by_symbol_since_limit(orders, symbol, since, limit, True)

    def handle_orders(self, client: Client, message):
        # spot update
        # {
        #        "closed":[
        #           {
        #              "action":"New",
        #              "avgPriceEp":4138763000000,
        #              "baseCurrency":"BTC",
        #              "baseQtyEv":0,
        #              "bizError":0,
        #              "clOrdID":"7956e0be-e8be-93a0-2887-ca504d85cda2",
        #              "createTimeNs":"1650442617606017583",
        #              "cumBaseQtyEv":30100,
        #              "cumFeeEv":31,
        #              "cumQuoteQtyEv":1245767663,
        #              "cxlRejReason":0,
        #              "feeCurrency":"BTC",
        #              "leavesBaseQtyEv":0,
        #              "leavesQuoteQtyEv":0,
        #              "ordStatus":"Filled",
        #              "ordType":"Market",
        #              "orderID":"34a4b1a8-ac3a-4580-b3e6-a6d039f27195",
        #              "pegOffsetValueEp":0,
        #              "priceEp":4549022000000,
        #              "qtyType":"ByQuote",
        #              "quoteCurrency":"USDT",
        #              "quoteQtyEv":1248000000,
        #              "side":"Buy",
        #              "stopPxEp":0,
        #              "symbol":"sBTCUSDT",
        #              "timeInForce":"ImmediateOrCancel",
        #              "tradeType":"Trade",
        #              "transactTimeNs":"1650442617609928764",
        #              "triggerTimeNs":0,
        #              "userID":2647224
        #           }
        #        ],
        #        "fills":[
        #           {
        #              "avgPriceEp":4138763000000,
        #              "baseCurrency":"BTC",
        #              "baseQtyEv":0,
        #              "clOrdID":"7956e0be-e8be-93a0-2887-ca504d85cda2",
        #              "execBaseQtyEv":30100,
        #              "execFeeEv":31,
        #              "execID":"d3b10cfa-84e3-5752-828e-78a79617e598",
        #              "execPriceEp":4138763000000,
        #              "execQuoteQtyEv":1245767663,
        #              "feeCurrency":"BTC",
        #              "lastLiquidityInd":"RemovedLiquidity",
        #              "ordType":"Market",
        #              "orderID":"34a4b1a8-ac3a-4580-b3e6-a6d039f27195",
        #              "priceEp":4549022000000,
        #              "qtyType":"ByQuote",
        #              "quoteCurrency":"USDT",
        #              "quoteQtyEv":1248000000,
        #              "side":"Buy",
        #              "symbol":"sBTCUSDT",
        #              "tradeType":"Trade",
        #              "transactTimeNs":"1650442617609928764",
        #              "userID":2647224
        #           }
        #        ],
        #        "open":[
        #           {
        #              "action":"New",
        #              "avgPriceEp":0,
        #              "baseCurrency":"LTC",
        #              "baseQtyEv":0,
        #              "bizError":0,
        #              "clOrdID":"2c0e5eb5-efb7-60d3-2e5f-df175df412ef",
        #              "createTimeNs":"1650446670073853755",
        #              "cumBaseQtyEv":0,
        #              "cumFeeEv":0,
        #              "cumQuoteQtyEv":0,
        #              "cxlRejReason":0,
        #              "feeCurrency":"LTC",
        #              "leavesBaseQtyEv":0,
        #              "leavesQuoteQtyEv":1000000000,
        #              "ordStatus":"New",
        #              "ordType":"Limit",
        #              "orderID":"d2aad92f-50f5-441a-957b-8184b146e3fb",
        #              "pegOffsetValueEp":0,
        #              "priceEp":5000000000,
        #              "qtyType":"ByQuote",
        #              "quoteCurrency":"USDT",
        #              "quoteQtyEv":1000000000,
        #              "side":"Buy",
        #            }
        #        ]
        #  },
        # perpetual
        #    [
        #        {
        #          "accountID": 40183400003,
        #          "action": "New",
        #          "actionBy": "ByUser",
        #          "actionTimeNs": "1674110665380190869",
        #          "addedSeq": 678760103,
        #          "apRp": "0",
        #          "bonusChangedAmountRv": "0",
        #          "bpRp": "0",
        #          "clOrdID": '',
        #          "cl_req_code": 0,
        #          "closedPnlRv": "0",
        #          "closedSize": "0",
        #          "code": 0,
        #          "cumFeeRv": "0",
        #          "cumQty": "0.001",
        #          "cumValueRv": "20.849",
        #          "curAccBalanceRv": "19.9874906",
        #          "curAssignedPosBalanceRv": "0",
        #          "curBonusBalanceRv": "0",
        #          "curLeverageRr": "-10",
        #          "curPosSide": "Buy",
        #          "curPosSize": "0.001",
        #          "curPosTerm": 1,
        #          "curPosValueRv": "20.849",
        #          "curRiskLimitRv": "1000000",
        #          "currency": "USDT",
        #          "cxlRejReason": 0,
        #          "displayQty": "0.001",
        #          "execFeeRv": "0.0125094",
        #          "execID": "b88d2950-04a2-52d8-8927-346059900242",
        #          "execPriceRp": "20849",
        #          "execQty": "0.001",
        #          "execSeq": 678760103,
        #          "execStatus": "TakerFill",
        #          "execValueRv": "20.849",
        #          "feeRateRr": "0.0006",
        #          "lastLiquidityInd": "RemovedLiquidity",
        #          "leavesQty": "0",
        #          "leavesValueRv": "0",
        #          "message": "No error",
        #          "ordStatus": "Filled",
        #          "ordType": "Market",
        #          "orderID": "79620ed2-54c6-4645-a35c-7057e687c576",
        #          "orderQty": "0.001",
        #          "pegOffsetProportionRr": "0",
        #          "pegOffsetValueRp": "0",
        #          "posSide": "Long",
        #          "priceRp": "21476.3",
        #          "relatedPosTerm": 1,
        #          "relatedReqNum": 4,
        #          "side": "Buy",
        #          "slTrigger": "ByMarkPrice",
        #          "stopLossRp": "0",
        #          "stopPxRp": "0",
        #          "symbol": "BTCUSDT",
        #          "takeProfitRp": "0",
        #          "timeInForce": "ImmediateOrCancel",
        #          "tpTrigger": "ByLastPrice",
        #          "tradeType": "Trade",
        #          "transactTimeNs": "1674110665387882268",
        #          "userID": 4018340
        #        },
        #        ...
        #    ]
        #
        trades = []
        parsedOrders = []
        if ('closed' in message) or ('fills' in message) or ('open' in message):
            closed = self.safe_value(message, 'closed', [])
            open = self.safe_value(message, 'open', [])
            orders = self.array_concat(open, closed)
            ordersLength = len(orders)
            if ordersLength == 0:
                return
            trades = self.safe_value(message, 'fills', [])
            for i in range(0, len(orders)):
                rawOrder = orders[i]
                parsedOrder = self.parse_order(rawOrder)
                parsedOrders.append(parsedOrder)
        else:
            for i in range(0, len(message)):
                update = message[i]
                action = self.safe_string(update, 'action')
                if (action is not None) and (action != 'Cancel'):
                    # order + trade info together
                    trades.append(update)
                parsedOrder = self.parse_ws_swap_order(update)
                parsedOrders.append(parsedOrder)
        self.handle_my_trades(client, trades)
        limit = self.safe_integer(self.options, 'ordersLimit', 1000)
        marketIds: dict = {}
        if self.orders is None:
            self.orders = ArrayCacheBySymbolById(limit)
        type = None
        stored = self.orders
        for i in range(0, len(parsedOrders)):
            parsed = parsedOrders[i]
            stored.append(parsed)
            symbol = parsed['symbol']
            market = self.market(symbol)
            if type is None:
                isUsdt = market['settle'] == 'USDT'
                type = 'perpetual' if isUsdt else market['type']
            marketIds[symbol] = True
        keys = list(marketIds.keys())
        for i in range(0, len(keys)):
            currentMessageHash = 'orders' + ':' + keys[i]
            client.resolve(self.orders, currentMessageHash)
        # resolve generic subscription(spot or swap)
        messageHash = 'orders:' + type
        client.resolve(self.orders, messageHash)

    def parse_ws_swap_order(self, order, market=None):
        #
        # swap
        #    {
        #        "accountID":26472240002,
        #        "action":"Cancel",
        #        "actionBy":"ByUser",
        #        "actionTimeNs":"1650450096104760797",
        #        "addedSeq":26975849309,
        #        "bonusChangedAmountEv":0,
        #        "clOrdID":"d9675963-5e4e-6fc8-898a-ec8b934c1c61",
        #        "closedPnlEv":0,
        #        "closedSize":0,
        #        "code":0,
        #        "cumQty":0,
        #        "cumValueEv":0,
        #        "curAccBalanceEv":400079,
        #        "curAssignedPosBalanceEv":0,
        #        "curBonusBalanceEv":0,
        #        "curLeverageEr":0,
        #        "curPosSide":"None",
        #        "curPosSize":0,
        #        "curPosTerm":1,
        #        "curPosValueEv":0,
        #        "curRiskLimitEv":5000000000,
        #        "currency":"USD",
        #        "cxlRejReason":0,
        #        "displayQty":0,
        #        "execFeeEv":0,
        #        "execID":"00000000-0000-0000-0000-000000000000",
        #        "execPriceEp":0,
        #        "execQty":1,
        #        "execSeq":26975862338,
        #        "execStatus":"Canceled",
        #        "execValueEv":0,
        #        "feeRateEr":0,
        #        "leavesQty":0,
        #        "leavesValueEv":0,
        #        "message":"No error",
        #        "ordStatus":"Canceled",
        #        "ordType":"Limit",
        #        "orderID":"8141deb9-8f94-48f6-9421-a4e3a791537b",
        #        "orderQty":1,
        #        "pegOffsetValueEp":0,
        #        "priceEp":9521,
        #        "relatedPosTerm":1,
        #        "relatedReqNum":4,
        #        "side":"Buy",
        #        "slTrigger":"ByMarkPrice",
        #        "stopLossEp":0,
        #        "stopPxEp":0,
        #        "symbol":"ADAUSD",
        #        "takeProfitEp":0,
        #        "timeInForce":"GoodTillCancel",
        #        "tpTrigger":"ByLastPrice",
        #        "transactTimeNs":"1650450096108143014",
        #        "userID":2647224
        #    }
        # perpetual
        #    {
        #        "accountID": 40183400003,
        #        "action": "New",
        #        "actionBy": "ByUser",
        #        "actionTimeNs": "1674110665380190869",
        #        "addedSeq": 678760103,
        #        "apRp": "0",
        #        "bonusChangedAmountRv": "0",
        #        "bpRp": "0",
        #        "clOrdID": '',
        #        "cl_req_code": 0,
        #        "closedPnlRv": "0",
        #        "closedSize": "0",
        #        "code": 0,
        #        "cumFeeRv": "0",
        #        "cumQty": "0.001",
        #        "cumValueRv": "20.849",
        #        "curAccBalanceRv": "19.9874906",
        #        "curAssignedPosBalanceRv": "0",
        #        "curBonusBalanceRv": "0",
        #        "curLeverageRr": "-10",
        #        "curPosSide": "Buy",
        #        "curPosSize": "0.001",
        #        "curPosTerm": 1,
        #        "curPosValueRv": "20.849",
        #        "curRiskLimitRv": "1000000",
        #        "currency": "USDT",
        #        "cxlRejReason": 0,
        #        "displayQty": "0.001",
        #        "execFeeRv": "0.0125094",
        #        "execID": "b88d2950-04a2-52d8-8927-346059900242",
        #        "execPriceRp": "20849",
        #        "execQty": "0.001",
        #        "execSeq": 678760103,
        #        "execStatus": "TakerFill",
        #        "execValueRv": "20.849",
        #        "feeRateRr": "0.0006",
        #        "lastLiquidityInd": "RemovedLiquidity",
        #        "leavesQty": "0",
        #        "leavesValueRv": "0",
        #        "message": "No error",
        #        "ordStatus": "Filled",
        #        "ordType": "Market",
        #        "orderID": "79620ed2-54c6-4645-a35c-7057e687c576",
        #        "orderQty": "0.001",
        #        "pegOffsetProportionRr": "0",
        #        "pegOffsetValueRp": "0",
        #        "posSide": "Long",
        #        "priceRp": "21476.3",
        #        "relatedPosTerm": 1,
        #        "relatedReqNum": 4,
        #        "side": "Buy",
        #        "slTrigger": "ByMarkPrice",
        #        "stopLossRp": "0",
        #        "stopPxRp": "0",
        #        "symbol": "BTCUSDT",
        #        "takeProfitRp": "0",
        #        "timeInForce": "ImmediateOrCancel",
        #        "tpTrigger": "ByLastPrice",
        #        "tradeType": "Trade",
        #        "transactTimeNs": "1674110665387882268",
        #        "userID": 4018340
        #    }
        #
        id = self.safe_string(order, 'orderID')
        clientOrderId = self.safe_string(order, 'clOrdID')
        if (clientOrderId is not None) and (len(clientOrderId) < 1):
            clientOrderId = None
        marketId = self.safe_string(order, 'symbol')
        market = self.safe_market(marketId, market)
        symbol = market['symbol']
        status = self.parse_order_status(self.safe_string(order, 'ordStatus'))
        side = self.safe_string_lower(order, 'side')
        type = self.parseOrderType(self.safe_string(order, 'ordType'))
        price = self.safe_string(order, 'priceRp', self.from_ep(self.safe_string(order, 'priceEp'), market))
        amount = self.safe_string(order, 'orderQty')
        filled = self.safe_string(order, 'cumQty')
        remaining = self.safe_string(order, 'leavesQty')
        timestamp = self.safe_integer_product(order, 'actionTimeNs', 0.000001)
        cost = self.safe_string(order, 'cumValueRv', self.from_ev(self.safe_string(order, 'cumValueEv'), market))
        lastTradeTimestamp = self.safe_integer_product(order, 'transactTimeNs', 0.000001)
        if lastTradeTimestamp == 0:
            lastTradeTimestamp = None
        timeInForce = self.parse_time_in_force(self.safe_string(order, 'timeInForce'))
        stopPrice = self.safe_string(order, 'stopPx')
        postOnly = (timeInForce == 'PO')
        return self.safe_order({
            'info': order,
            'id': id,
            'clientOrderId': clientOrderId,
            'datetime': self.iso8601(timestamp),
            'timestamp': timestamp,
            'lastTradeTimestamp': lastTradeTimestamp,
            'symbol': symbol,
            'type': type,
            'timeInForce': timeInForce,
            'postOnly': postOnly,
            'side': side,
            'price': price,
            'stopPrice': stopPrice,
            'triggerPrice': stopPrice,
            'amount': amount,
            'filled': filled,
            'remaining': remaining,
            'cost': cost,
            'average': None,
            'status': status,
            'fee': None,
            'trades': None,
        }, market)

    def handle_message(self, client: Client, message):
        # private spot update
        # {
        #     "orders": {closed: [], fills: [], open: []},
        #     "sequence": 40435835,
        #     "timestamp": "1650443245600839241",
        #     "type": "snapshot",
        #     "wallets": [
        #       {
        #         "balanceEv": 0,
        #         "currency": "BTC",
        #         "lastUpdateTimeNs": "1650442638722099092",
        #         "lockedTradingBalanceEv": 0,
        #         "lockedWithdrawEv": 0,
        #         "userID": 2647224
        #       },
        #       {
        #         "balanceEv": 1154232337,
        #         "currency": "USDT",
        #         "lastUpdateTimeNs": "1650442617610017597",
        #         "lockedTradingBalanceEv": 0,
        #         "lockedWithdrawEv": 0,
        #         "userID": 2647224
        #       }
        #     ]
        # }
        # private swap update
        # {
        #     "sequence": 83839628,
        #     "timestamp": "1650382581827447829",
        #     "type": "snapshot",
        #     "accounts": [
        #       {
        #         "accountBalanceEv": 0,
        #         "accountID": 26472240001,
        #         "bonusBalanceEv": 0,
        #         "currency": "BTC",
        #         "totalUsedBalanceEv": 0,
        #         "userID": 2647224
        #       }
        #     ],
        #     "orders": [],
        #     "positions": [
        #       {
        #         "accountID": 26472240001,
        #         "assignedPosBalanceEv": 0,
        #         "avgEntryPriceEp": 0,
        #         "bankruptCommEv": 0,
        #         "bankruptPriceEp": 0,
        #         "buyLeavesQty": 0,
        #         "buyLeavesValueEv": 0,
        #         "buyValueToCostEr": 1150750,
        #         "createdAtNs": 0,
        #         "crossSharedBalanceEv": 0,
        #         "cumClosedPnlEv": 0,
        #         "cumFundingFeeEv": 0,
        #         "cumTransactFeeEv": 0,
        #         "curTermRealisedPnlEv": 0,
        #         "currency": "BTC",
        #         "dataVer": 2,
        #         "deleveragePercentileEr": 0,
        #         "displayLeverageEr": 10000000000,
        #         "estimatedOrdLossEv": 0,
        #         "execSeq": 0,
        #         "freeCostEv": 0,
        #         "freeQty": 0,
        #         "initMarginReqEr": 1000000,
        #         "lastFundingTime": "1640601827712091793",
        #         "lastTermEndTime": 0,
        #         "leverageEr": 0,
        #         "liquidationPriceEp": 0,
        #         "maintMarginReqEr": 500000,
        #         "makerFeeRateEr": 0,
        #         "markPriceEp": 507806777,
        #         "orderCostEv": 0,
        #         "posCostEv": 0,
        #         "positionMarginEv": 0,
        #         "positionStatus": "Normal",
        #         "riskLimitEv": 10000000000,
        #         "sellLeavesQty": 0,
        #         "sellLeavesValueEv": 0,
        #         "sellValueToCostEr": 1149250,
        #         "side": "None",
        #         "size": 0,
        #         "symbol": "BTCUSD",
        #         "takerFeeRateEr": 0,
        #         "term": 1,
        #         "transactTimeNs": 0,
        #         "unrealisedPnlEv": 0,
        #         "updatedAtNs": 0,
        #         "usedBalanceEv": 0,
        #         "userID": 2647224,
        #         "valueEv": 0
        #       }
        #     ]
        # }
        id = self.safe_string(message, 'id')
        if id in client.subscriptions:
            method = client.subscriptions[id]
            del client.subscriptions[id]
            if method is not True:
                method(client, message)
                return
        methodName = self.safe_string(message, 'method', '')
        if ('market24h' in message) or ('spot_market24h' in message) or (methodName.find('perp_market24h_pack_p') >= 0):
            self.handle_ticker(client, message)
            return
        elif ('trades' in message) or ('trades_p' in message):
            self.handle_trades(client, message)
            return
        elif ('kline' in message) or ('kline_p' in message):
            self.handle_ohlcv(client, message)
            return
        elif ('book' in message) or ('orderbook_p' in message):
            self.handle_order_book(client, message)
            return
        if ('orders' in message) or ('orders_p' in message):
            orders = self.safe_value_2(message, 'orders', 'orders_p', {})
            self.handle_orders(client, orders)
        if ('accounts' in message) or ('accounts_p' in message) or ('wallets' in message):
            type = 'swap' if ('accounts' in message) else 'spot'
            if 'accounts_p' in message:
                type = 'perpetual'
            accounts = self.safe_value_n(message, ['accounts', 'accounts_p', 'wallets'], [])
            self.handle_balance(type, client, accounts)

    def handle_authenticate(self, client: Client, message):
        #
        # {
        #     "error": null,
        #     "id": 1234,
        #     "result": {
        #       "status": "success"
        #     }
        # }
        #
        result = self.safe_value(message, 'result')
        status = self.safe_string(result, 'status')
        messageHash = 'authenticated'
        if status == 'success':
            client.resolve(message, messageHash)
        else:
            error = AuthenticationError(self.id + ' ' + self.json(message))
            client.reject(error, messageHash)
            if messageHash in client.subscriptions:
                del client.subscriptions[messageHash]

    async def subscribe_private(self, type, messageHash, params={}):
        await self.load_markets()
        await self.authenticate()
        url = self.urls['api']['ws']
        requestId = self.seconds()
        settleIsUSDT = (self.safe_value(params, 'settle', '') == 'USDT')
        params = self.omit(params, 'settle')
        channel = 'aop.subscribe'
        if type == 'spot':
            channel = 'wo.subscribe'
        if settleIsUSDT:
            channel = 'aop_p.subscribe'
        request = {
            'id': requestId,
            'method': channel,
            'params': [],
        }
        request = self.extend(request, params)
        return await self.watch(url, messageHash, request, channel)

    async def authenticate(self, params={}):
        self.check_required_credentials()
        url = self.urls['api']['ws']
        client = self.client(url)
        requestId = self.request_id()
        messageHash = 'authenticated'
        future = self.safe_value(client.subscriptions, messageHash)
        if future is None:
            expiryDelta = self.safe_integer(self.options, 'expires', 120)
            expiration = self.seconds() + expiryDelta
            payload = self.apiKey + str(expiration)
            signature = self.hmac(self.encode(payload), self.encode(self.secret), hashlib.sha256)
            method = 'user.auth'
            request: dict = {
                'method': method,
                'params': ['API', self.apiKey, signature, expiration],
                'id': requestId,
            }
            subscriptionHash = str(requestId)
            message = self.extend(request, params)
            if not (messageHash in client.subscriptions):
                client.subscriptions[subscriptionHash] = self.handle_authenticate
            future = await self.watch(url, messageHash, message, messageHash)
            client.subscriptions[messageHash] = future
        return future




# ----------------------------------------------------------------------------
# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code
# -----------------------------------------------------------------------------

from ccxt.test.base.test_account import test_account # noqa E402
from ccxt.test.base.test_balance import test_balance # noqa E402
from ccxt.test.base.test_borrow_interest import test_borrow_interest # noqa E402
from ccxt.test.base.test_borrow_rate import test_borrow_rate # noqa E402
from ccxt.test.base.test_currency import test_currency # noqa E402
from ccxt.test.base.test_deposit_withdrawal import test_deposit_withdrawal # noqa E402
from ccxt.test.base.test_funding_rate_history import test_funding_rate_history # noqa E402
from ccxt.test.base.test_last_price import test_last_price # noqa E402
from ccxt.test.base.test_ledger_entry import test_ledger_entry # noqa E402
from ccxt.test.base.test_leverage_tier import test_leverage_tier # noqa E402
from ccxt.test.base.test_liquidation import test_liquidation # noqa E402
from ccxt.test.base.test_margin_mode import test_margin_mode # noqa E402
from ccxt.test.base.test_margin_modification import test_margin_modification # noqa E402
from ccxt.test.base.test_market import test_market # noqa E402
from ccxt.test.base.test_ohlcv import test_ohlcv # noqa E402
from ccxt.test.base.test_open_interest import test_open_interest # noqa E402
from ccxt.test.base.test_order import test_order # noqa E402
from ccxt.test.base.test_order_book import test_order_book # noqa E402
from ccxt.test.base.test_position import test_position # noqa E402
from ccxt.test.base.test_status import test_status # noqa E402
from ccxt.test.base.test_ticker import test_ticker # noqa E402
from ccxt.test.base.test_trade import test_trade # noqa E402
from ccxt.test.base.test_trading_fee import test_trading_fee # noqa E402


import os
import sys

root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
sys.path.append(root)

# ----------------------------------------------------------------------------

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

# ----------------------------------------------------------------------------

from ccxt.base.decimal_to_precision import decimal_to_precision  # noqa F401
from ccxt.base.decimal_to_precision import TRUNCATE              # noqa F401
from ccxt.base.decimal_to_precision import ROUND                 # noqa F401
from ccxt.base.decimal_to_precision import DECIMAL_PLACES        # noqa F401
from ccxt.base.decimal_to_precision import SIGNIFICANT_DIGITS    # noqa F401
from ccxt.base.decimal_to_precision import TICK_SIZE             # noqa F401
from ccxt.base.decimal_to_precision import PAD_WITH_ZERO         # noqa F401
from ccxt.base.decimal_to_precision import NO_PADDING            # noqa F401
from ccxt.base.decimal_to_precision import number_to_string      # noqa F401
from ccxt.base.exchange import Exchange                          # noqa F401
from ccxt.base.precise import Precise                            # noqa F401

# eslint-disable-next-line import/newline-after-import

# ----------------------------------------------------------------------------
# number_to_string

assert number_to_string(-7.8e-7) == '-0.00000078'
assert number_to_string(7.8e-7) == '0.00000078'
assert number_to_string(-17.805e-7) == '-0.0000017805'
assert number_to_string(17.805e-7) == '0.0000017805'
assert number_to_string(-7.0005e27) == '-7000500000000000000000000000'
assert number_to_string(7.0005e27) == '7000500000000000000000000000'
assert number_to_string(-7.9e27) == '-7900000000000000000000000000'
assert number_to_string(7e27) == '7000000000000000000000000000'
assert number_to_string(7.9e27) == '7900000000000000000000000000'
assert number_to_string(-12.345) == '-12.345'
assert number_to_string(12.345) == '12.345'
assert number_to_string(0) == '0'
assert number_to_string(7.35946e21) == '7359460000000000000000'
assert number_to_string(0.00000001) == '0.00000001'
assert number_to_string(1e-7) == '0.0000001'
assert number_to_string(-1e-7) == '-0.0000001'

# ----------------------------------------------------------------------------
# testDecimalToPrecisionTruncationToNDigitsAfterDot

assert decimal_to_precision('12.3456000', TRUNCATE, 100, DECIMAL_PLACES) == '12.3456'
assert decimal_to_precision('12.3456', TRUNCATE, 100, DECIMAL_PLACES) == '12.3456'
assert decimal_to_precision('12.3456', TRUNCATE, 4, DECIMAL_PLACES) == '12.3456'
assert decimal_to_precision('12.3456', TRUNCATE, 3, DECIMAL_PLACES) == '12.345'
assert decimal_to_precision('12.3456', TRUNCATE, 2, DECIMAL_PLACES) == '12.34'
assert decimal_to_precision('12.3456', TRUNCATE, 1, DECIMAL_PLACES) == '12.3'
assert decimal_to_precision('12.3456', TRUNCATE, 0, DECIMAL_PLACES) == '12'
# ['12.3456',    TRUNCATE,  -1, DECIMAL_PLACES,  '10'],   # not yet supported
# ['123.456',    TRUNCATE,  -2, DECIMAL_PLACES,  '120'],  # not yet supported
# ['123.456',    TRUNCATE,  -3, DECIMAL_PLACES,  '100'],  # not yet supported

assert decimal_to_precision('0.0000001', TRUNCATE, 8, DECIMAL_PLACES) == '0.0000001'
assert decimal_to_precision('0.00000001', TRUNCATE, 8, DECIMAL_PLACES) == '0.00000001'

assert decimal_to_precision('0.000000000', TRUNCATE, 9, DECIMAL_PLACES, PAD_WITH_ZERO) == '0.000000000'
assert decimal_to_precision('0.000000001', TRUNCATE, 9, DECIMAL_PLACES, PAD_WITH_ZERO) == '0.000000001'

assert decimal_to_precision('12.3456', TRUNCATE, -1, DECIMAL_PLACES) == '10'
assert decimal_to_precision('123.456', TRUNCATE, -1, DECIMAL_PLACES) == '120'
assert decimal_to_precision('123.456', TRUNCATE, -2, DECIMAL_PLACES) == '100'
assert decimal_to_precision('9.99999', TRUNCATE, -1, DECIMAL_PLACES) == '0'
assert decimal_to_precision('99.9999', TRUNCATE, -1, DECIMAL_PLACES) == '90'
assert decimal_to_precision('99.9999', TRUNCATE, -2, DECIMAL_PLACES) == '0'

assert decimal_to_precision('0', TRUNCATE, 0, DECIMAL_PLACES) == '0'
assert decimal_to_precision('-0.9', TRUNCATE, 0, DECIMAL_PLACES) == '0'

# ----------------------------------------------------------------------------
# testDecimalToPrecisionTruncationToNSignificantDigits

assert decimal_to_precision('0.000123456700', TRUNCATE, 100, SIGNIFICANT_DIGITS) == '0.0001234567'
assert decimal_to_precision('0.0001234567', TRUNCATE, 100, SIGNIFICANT_DIGITS) == '0.0001234567'
assert decimal_to_precision('0.0001234567', TRUNCATE, 7, SIGNIFICANT_DIGITS) == '0.0001234567'

assert decimal_to_precision('0.000123456', TRUNCATE, 6, SIGNIFICANT_DIGITS) == '0.000123456'
assert decimal_to_precision('0.000123456', TRUNCATE, 5, SIGNIFICANT_DIGITS) == '0.00012345'
assert decimal_to_precision('0.000123456', TRUNCATE, 2, SIGNIFICANT_DIGITS) == '0.00012'
assert decimal_to_precision('0.000123456', TRUNCATE, 1, SIGNIFICANT_DIGITS) == '0.0001'

assert decimal_to_precision('123.0000987654', TRUNCATE, 10, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '123.0000987'
assert decimal_to_precision('123.0000987654', TRUNCATE, 8, SIGNIFICANT_DIGITS) == '123.00009'
assert decimal_to_precision('123.0000987654', TRUNCATE, 7, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '123.0000'
assert decimal_to_precision('123.0000987654', TRUNCATE, 6, SIGNIFICANT_DIGITS) == '123'
assert decimal_to_precision('123.0000987654', TRUNCATE, 5, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '123.00'
assert decimal_to_precision('123.0000987654', TRUNCATE, 4, SIGNIFICANT_DIGITS) == '123'
assert decimal_to_precision('123.0000987654', TRUNCATE, 4, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '123.0'
assert decimal_to_precision('123.0000987654', TRUNCATE, 3, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '123'
assert decimal_to_precision('123.0000987654', TRUNCATE, 2, SIGNIFICANT_DIGITS) == '120'
assert decimal_to_precision('123.0000987654', TRUNCATE, 1, SIGNIFICANT_DIGITS) == '100'
assert decimal_to_precision('123.0000987654', TRUNCATE, 1, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '100'

assert decimal_to_precision('1234', TRUNCATE, 5, SIGNIFICANT_DIGITS) == '1234'
assert decimal_to_precision('1234', TRUNCATE, 5, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '1234.0'
assert decimal_to_precision('1234', TRUNCATE, 4, SIGNIFICANT_DIGITS) == '1234'
assert decimal_to_precision('1234', TRUNCATE, 4, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '1234'
assert decimal_to_precision('1234.69', TRUNCATE, 0, SIGNIFICANT_DIGITS) == '0'
assert decimal_to_precision('1234.69', TRUNCATE, 0, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '0'

# ----------------------------------------------------------------------------
# testDecimalToPrecisionRoundingToNDigitsAfterDot

assert decimal_to_precision('12.3456000', ROUND, 100, DECIMAL_PLACES) == '12.3456'
assert decimal_to_precision('12.3456', ROUND, 100, DECIMAL_PLACES) == '12.3456'
assert decimal_to_precision('12.3456', ROUND, 4, DECIMAL_PLACES) == '12.3456'
assert decimal_to_precision('12.3456', ROUND, 3, DECIMAL_PLACES) == '12.346'
assert decimal_to_precision('12.3456', ROUND, 2, DECIMAL_PLACES) == '12.35'
assert decimal_to_precision('12.3456', ROUND, 1, DECIMAL_PLACES) == '12.3'
assert decimal_to_precision('12.3456', ROUND, 0, DECIMAL_PLACES) == '12'

# todo:
# ['9.999',     ROUND,   3, DECIMAL_PLACES,    NO_PADDING,  '9.999'],
# ['9.999',     ROUND,   2, DECIMAL_PLACES,    NO_PADDING,  '10'],
# ['9.999',     ROUND,   2, DECIMAL_PLACES, PAD_WITH_ZERO,  '10.00'],
# ['99.999',    ROUND,   2, DECIMAL_PLACES, PAD_WITH_ZERO,  '100.00'],
# ['-99.999',    ROUND,   2, DECIMAL_PLACES, PAD_WITH_ZERO, '-100.00'],

# ['12.3456',    ROUND,  -1, DECIMAL_PLACES,    NO_PADDING,  '10'],  # not yet supported
# ['123.456',    ROUND,  -1, DECIMAL_PLACES,    NO_PADDING,  '120'],  # not yet supported
# ['123.456',    ROUND,  -2, DECIMAL_PLACES,    NO_PADDING,  '100'],  # not yet supported

# a problematic case in PHP
assert decimal_to_precision('10000', ROUND, 6, DECIMAL_PLACES) == '10000'
assert decimal_to_precision('0.00003186', ROUND, 8, DECIMAL_PLACES) == '0.00003186'

assert decimal_to_precision('12.3456', ROUND, -1, DECIMAL_PLACES) == '10'
assert decimal_to_precision('123.456', ROUND, -1, DECIMAL_PLACES) == '120'
assert decimal_to_precision('123.456', ROUND, -2, DECIMAL_PLACES) == '100'
assert decimal_to_precision('9.99999', ROUND, -1, DECIMAL_PLACES) == '10'
assert decimal_to_precision('99.9999', ROUND, -1, DECIMAL_PLACES) == '100'
assert decimal_to_precision('99.9999', ROUND, -2, DECIMAL_PLACES) == '100'

assert decimal_to_precision('9.999', ROUND, 3, DECIMAL_PLACES) == '9.999'
assert decimal_to_precision('9.999', ROUND, 2, DECIMAL_PLACES) == '10'
assert decimal_to_precision('9.999', ROUND, 2, DECIMAL_PLACES, PAD_WITH_ZERO) == '10.00'
assert decimal_to_precision('99.999', ROUND, 2, DECIMAL_PLACES, PAD_WITH_ZERO) == '100.00'
assert decimal_to_precision('-99.999', ROUND, 2, DECIMAL_PLACES, PAD_WITH_ZERO) == '-100.00'

# ----------------------------------------------------------------------------
# testDecimalToPrecisionRoundingToNSignificantDigits

assert decimal_to_precision('0.000123456700', ROUND, 100, SIGNIFICANT_DIGITS) == '0.0001234567'
assert decimal_to_precision('0.0001234567', ROUND, 100, SIGNIFICANT_DIGITS) == '0.0001234567'
assert decimal_to_precision('0.0001234567', ROUND, 7, SIGNIFICANT_DIGITS) == '0.0001234567'

assert decimal_to_precision('0.000123456', ROUND, 6, SIGNIFICANT_DIGITS) == '0.000123456'
assert decimal_to_precision('0.000123456', ROUND, 5, SIGNIFICANT_DIGITS) == '0.00012346'
assert decimal_to_precision('0.000123456', ROUND, 4, SIGNIFICANT_DIGITS) == '0.0001235'
assert decimal_to_precision('0.00012', ROUND, 2, SIGNIFICANT_DIGITS) == '0.00012'
assert decimal_to_precision('0.0001', ROUND, 1, SIGNIFICANT_DIGITS) == '0.0001'

assert decimal_to_precision('123.0000987654', ROUND, 7, SIGNIFICANT_DIGITS) == '123.0001'
assert decimal_to_precision('123.0000987654', ROUND, 6, SIGNIFICANT_DIGITS) == '123'

assert decimal_to_precision('0.00098765', ROUND, 2, SIGNIFICANT_DIGITS) == '0.00099'
assert decimal_to_precision('0.00098765', ROUND, 2, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '0.00099'

assert decimal_to_precision('0.00098765', ROUND, 1, SIGNIFICANT_DIGITS) == '0.001'
assert decimal_to_precision('0.00098765', ROUND, 10, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '0.0009876500000'

assert decimal_to_precision('0.098765', ROUND, 1, SIGNIFICANT_DIGITS, PAD_WITH_ZERO) == '0.1'

assert decimal_to_precision('0', ROUND, 0, SIGNIFICANT_DIGITS) == '0'
assert decimal_to_precision('-0.123', ROUND, 0, SIGNIFICANT_DIGITS) == '0'

assert decimal_to_precision('0.00000044', ROUND, 5, SIGNIFICANT_DIGITS) == '0.00000044'

# ----------------------------------------------------------------------------
# testDecimalToPrecisionRoundingToTickSize

assert decimal_to_precision('0.000123456700', ROUND, 0.00012, TICK_SIZE) == '0.00012'
assert decimal_to_precision('0.0001234567', ROUND, 0.00013, TICK_SIZE) == '0.00013'
assert decimal_to_precision('0.0001234567', TRUNCATE, 0.00013, TICK_SIZE) == '0'
assert decimal_to_precision('101.000123456700', ROUND, 100, TICK_SIZE) == '100'
assert decimal_to_precision('0.000123456700', ROUND, 100, TICK_SIZE) == '0'
assert decimal_to_precision('165', TRUNCATE, 110, TICK_SIZE) == '110'
assert decimal_to_precision('3210', TRUNCATE, 1110, TICK_SIZE) == '2220'
assert decimal_to_precision('165', ROUND, 110, TICK_SIZE) == '220'
assert decimal_to_precision('0.000123456789', ROUND, 0.00000012, TICK_SIZE) == '0.00012348'
assert decimal_to_precision('0.000123456789', TRUNCATE, 0.00000012, TICK_SIZE) == '0.00012336'
assert decimal_to_precision('0.000273398', ROUND, 1e-7, TICK_SIZE) == '0.0002734'

assert decimal_to_precision('0.00005714', TRUNCATE, 0.00000001, TICK_SIZE) == '0.00005714'
# self line causes problems in JS, fix with Precise
# assert decimal_to_precision('0.0000571495257361', TRUNCATE, 0.00000001, TICK_SIZE) == '0.00005714'

assert decimal_to_precision('0.01', ROUND, 0.0001, TICK_SIZE, PAD_WITH_ZERO) == '0.0100'
assert decimal_to_precision('0.01', TRUNCATE, 0.0001, TICK_SIZE, PAD_WITH_ZERO) == '0.0100'

assert decimal_to_precision('-0.000123456789', ROUND, 0.00000012, TICK_SIZE) == '-0.00012348'
assert decimal_to_precision('-0.000123456789', TRUNCATE, 0.00000012, TICK_SIZE) == '-0.00012336'
assert decimal_to_precision('-165', TRUNCATE, 110, TICK_SIZE) == '-110'
assert decimal_to_precision('-165', ROUND, 110, TICK_SIZE) == '-220'
assert decimal_to_precision('-1650', TRUNCATE, 1100, TICK_SIZE) == '-1100'
assert decimal_to_precision('-1650', ROUND, 1100, TICK_SIZE) == '-2200'

assert decimal_to_precision('0.0006', TRUNCATE, 0.0001, TICK_SIZE) == '0.0006'
assert decimal_to_precision('-0.0006', TRUNCATE, 0.0001, TICK_SIZE) == '-0.0006'
assert decimal_to_precision('0.6', TRUNCATE, 0.2, TICK_SIZE) == '0.6'
assert decimal_to_precision('-0.6', TRUNCATE, 0.2, TICK_SIZE) == '-0.6'
assert decimal_to_precision('1.2', ROUND, 0.4, TICK_SIZE) == '1.2'
assert decimal_to_precision('-1.2', ROUND, 0.4, TICK_SIZE) == '-1.2'
assert decimal_to_precision('1.2', ROUND, 0.02, TICK_SIZE) == '1.2'
assert decimal_to_precision('-1.2', ROUND, 0.02, TICK_SIZE) == '-1.2'
assert decimal_to_precision('44', ROUND, 4.4, TICK_SIZE) == '44'
assert decimal_to_precision('-44', ROUND, 4.4, TICK_SIZE) == '-44'
assert decimal_to_precision('44.00000001', ROUND, 4.4, TICK_SIZE) == '44'
assert decimal_to_precision('-44.00000001', ROUND, 4.4, TICK_SIZE) == '-44'

# https://github.com/ccxt/ccxt/issues/6731
assert decimal_to_precision('20', TRUNCATE, 0.00000001, TICK_SIZE) == '20'

# ----------------------------------------------------------------------------
# testDecimalToPrecisionNegativeNumbers

assert decimal_to_precision('-0.123456', TRUNCATE, 5, DECIMAL_PLACES) == '-0.12345'
assert decimal_to_precision('-0.123456', ROUND, 5, DECIMAL_PLACES) == '-0.12346'

# ----------------------------------------------------------------------------
# decimal_to_precision: without dot / trailing dot

assert decimal_to_precision('123', TRUNCATE, 0) == '123'

assert decimal_to_precision('123', TRUNCATE, 5, DECIMAL_PLACES) == '123'
assert decimal_to_precision('123', TRUNCATE, 5, DECIMAL_PLACES, PAD_WITH_ZERO) == '123.00000'

assert decimal_to_precision('123.', TRUNCATE, 0, DECIMAL_PLACES) == '123'
assert decimal_to_precision('123.', TRUNCATE, 5, DECIMAL_PLACES, PAD_WITH_ZERO) == '123.00000'

assert decimal_to_precision('0.', TRUNCATE, 0) == '0'
assert decimal_to_precision('0.', TRUNCATE, 5, DECIMAL_PLACES, PAD_WITH_ZERO) == '0.00000'

# ----------------------------------------------------------------------------
# decimal_to_precision: rounding for equidistant digits

assert decimal_to_precision('1.44', ROUND, 1, DECIMAL_PLACES) == '1.4'
assert decimal_to_precision('1.45', ROUND, 1, DECIMAL_PLACES) == '1.5'
assert decimal_to_precision('1.45', ROUND, 0, DECIMAL_PLACES) == '1'  # not 2

# ----------------------------------------------------------------------------
# negative precision only implemented so far in python
# pretty useless for decimal applications |x| < 5 == 0
# NO_PADDING and PAD_WITH_ZERO are ignored

assert decimal_to_precision('5', ROUND, -1, DECIMAL_PLACES) == '10'
assert decimal_to_precision('4.999', ROUND, -1, DECIMAL_PLACES) == '0'
assert decimal_to_precision('0.0431531423', ROUND, -1, DECIMAL_PLACES) == '0'
assert decimal_to_precision('-69.3', ROUND, -1, DECIMAL_PLACES) == '-70'
assert decimal_to_precision('5001', ROUND, -4, DECIMAL_PLACES) == '10000'
assert decimal_to_precision('4999.999', ROUND, -4, DECIMAL_PLACES) == '0'

assert decimal_to_precision('69.3', TRUNCATE, -2, DECIMAL_PLACES) == '0'
assert decimal_to_precision('-69.3', TRUNCATE, -2, DECIMAL_PLACES) == '0'
assert decimal_to_precision('69.3', TRUNCATE, -1, SIGNIFICANT_DIGITS) == '60'
assert decimal_to_precision('-69.3', TRUNCATE, -1, SIGNIFICANT_DIGITS) == '-60'
assert decimal_to_precision('69.3', TRUNCATE, -2, SIGNIFICANT_DIGITS) == '0'
assert decimal_to_precision('1602000000000000000000', TRUNCATE, 3, SIGNIFICANT_DIGITS) == '1600000000000000000000'

# ----------------------------------------------------------------------------
# decimal_to_precision: stringified precision
assert decimal_to_precision('-0.000123456789', ROUND, '0.00000012', TICK_SIZE) == '-0.00012348'
assert decimal_to_precision('-0.000123456789', TRUNCATE, '0.00000012', TICK_SIZE) == '-0.00012336'
assert decimal_to_precision('-165', TRUNCATE, '110', TICK_SIZE) == '-110'
assert decimal_to_precision('-165', ROUND, '110', TICK_SIZE) == '-220'

# ----------------------------------------------------------------------------
# testDecimalToPrecisionErrorHandling(todo)
#
# throws(() =>
#     decimal_to_precision('123456.789', TRUNCATE, -2, DECIMAL_PLACES),
#         'negative precision is not yet supported')
#
# throws(() =>
#     decimal_to_precision('foo'),
#         "invalid number(contains an illegal character 'f')")
#
# throws(() =>
#     decimal_to_precision('0.01', TRUNCATE, -1, TICK_SIZE),
#         "TICK_SIZE cant be used with negative numPrecisionDigits")

# ----------------------------------------------------------------------------

w = '-1.123e-6'
x = '0.00000002'
y = '69696900000'
z = '0'
a = '1e8'

assert Precise.string_mul(x, y) == '1393.938'
assert Precise.string_mul(y, x) == '1393.938'
assert Precise.string_add(x, y) == '69696900000.00000002'
assert Precise.string_add(y, x) == '69696900000.00000002'
assert Precise.string_sub(x, y) == '-69696899999.99999998'
assert Precise.string_sub(y, x) == '69696899999.99999998'
assert Precise.string_div(x, y, 1) == '0'
assert Precise.string_div(x, y) == '0'
assert Precise.string_div(x, y, 19) == '0.0000000000000000002'
assert Precise.string_div(x, y, 20) == '0.00000000000000000028'
assert Precise.string_div(x, y, 21) == '0.000000000000000000286'
assert Precise.string_div(x, y, 22) == '0.0000000000000000002869'
assert Precise.string_div(y, x) == '3484845000000000000'

assert Precise.string_mul(x, w) == '-0.00000000000002246'
assert Precise.string_mul(w, x) == '-0.00000000000002246'
assert Precise.string_add(x, w) == '-0.000001103'
assert Precise.string_add(w, x) == '-0.000001103'
assert Precise.string_sub(x, w) == '0.000001143'
assert Precise.string_sub(w, x) == '-0.000001143'
assert Precise.string_div(x, w) == '-0.017809439002671415'
assert Precise.string_div(w, x) == '-56.15'

assert Precise.string_mul(z, w) == '0'
assert Precise.string_mul(z, x) == '0'
assert Precise.string_mul(z, y) == '0'
assert Precise.string_mul(w, z) == '0'
assert Precise.string_mul(x, z) == '0'
assert Precise.string_mul(y, z) == '0'
assert Precise.string_add(z, w) == '-0.000001123'
assert Precise.string_add(z, x) == '0.00000002'
assert Precise.string_add(z, y) == '69696900000'
assert Precise.string_add(w, z) == '-0.000001123'
assert Precise.string_add(x, z) == '0.00000002'
assert Precise.string_add(y, z) == '69696900000'

assert Precise.string_mul(x, a) == '2'
assert Precise.string_mul(a, x) == '2'
assert Precise.string_mul(y, a) == '6969690000000000000'
assert Precise.string_mul(a, y) == '6969690000000000000'
assert Precise.string_div(y, a) == '696.969'
assert Precise.string_div(y, a, -1) == '690'
assert Precise.string_div(y, a, 0) == '696'
assert Precise.string_div(y, a, 1) == '696.9'
assert Precise.string_div(y, a, 2) == '696.96'
assert Precise.string_div(a, y) == '0.001434784043479695'

assert Precise.string_abs('0') == '0'
assert Precise.string_abs('-0') == '0'
assert Precise.string_abs('-500.1') == '500.1'
assert Precise.string_abs('213') == '213'

assert Precise.string_neg('0') == '0'
assert Precise.string_neg('-0') == '0'
assert Precise.string_neg('-500.1') == '500.1'
assert Precise.string_neg('213') == '-213'

assert Precise.string_mod('57.123', '10') == '7.123'
assert Precise.string_mod('18', '6') == '0'
assert Precise.string_mod('10.1', '0.5') == '0.1'
assert Precise.string_mod('10000000', '5555') == '1000'
assert Precise.string_mod('5550', '120') == '30'

assert Precise.string_equals('1.0000', '1')
assert Precise.string_equals('-0.0', '0')
assert Precise.string_equals('-0.0', '0.0')
assert Precise.string_equals('5.534000', '5.5340')

assert Precise.string_min('1.0000', '2') == '1'
assert Precise.string_min('2', '1.2345') == '1.2345'
assert Precise.string_min('3.1415', '-2') == '-2'
assert Precise.string_min('-3.1415', '-2') == '-3.1415'
assert Precise.string_min('0.000', '-0.0') == '0'

assert Precise.string_max('1.0000', '2') == '2'
assert Precise.string_max('2', '1.2345') == '2'
assert Precise.string_max('3.1415', '-2') == '3.1415'
assert Precise.string_max('-3.1415', '-2') == '-2'
assert Precise.string_max('0.000', '-0.0') == '0'

assert not Precise.string_gt('1.0000', '2')
assert Precise.string_gt('2', '1.2345')
assert Precise.string_gt('3.1415', '-2')
assert not Precise.string_gt('-3.1415', '-2')
assert not Precise.string_gt('3.1415', '3.1415')
assert Precise.string_gt('3.14150000000000000000001', '3.1415')

assert not Precise.string_ge('1.0000', '2')
assert Precise.string_ge('2', '1.2345')
assert Precise.string_ge('3.1415', '-2')
assert not Precise.string_ge('-3.1415', '-2')
assert Precise.string_ge('3.1415', '3.1415')
assert Precise.string_ge('3.14150000000000000000001', '3.1415')

assert Precise.string_lt('1.0000', '2')
assert not Precise.string_lt('2', '1.2345')
assert not Precise.string_lt('3.1415', '-2')
assert Precise.string_lt('-3.1415', '-2')
assert not Precise.string_lt('3.1415', '3.1415')
assert Precise.string_lt('3.1415', '3.14150000000000000000001')

assert Precise.string_le('1.0000', '2')
assert not Precise.string_le('2', '1.2345')
assert not Precise.string_le('3.1415', '-2')
assert Precise.string_le('-3.1415', '-2')
assert Precise.string_le('3.1415', '3.1415')
assert Precise.string_le('3.1415', '3.14150000000000000000001')

# todo
# $self->assertSame(0,   Exchange::sum())
# $self->assertSame(2,   Exchange::sum(2))
# $self->assertSame(432, Exchange::sum(2, 30, 400))
# eslint-disable-next-line eol-last
# $self->assertSame(439, Exchange::sum(2, null, [88], 30, '7', 400, null))



from ccxt.base.types import Entry


class ImplicitAPI:
    v1_public_get_assets = v1PublicGetAssets = Entry('assets', ['v1', 'public'], 'GET', {})
    v1_public_get_assets_assets = v1PublicGetAssetsAssets = Entry('assets/{assets}', ['v1', 'public'], 'GET', {})
    v1_public_get_assets_asset_networks = v1PublicGetAssetsAssetNetworks = Entry('assets/{asset}/networks', ['v1', 'public'], 'GET', {})
    v1_public_get_instruments = v1PublicGetInstruments = Entry('instruments', ['v1', 'public'], 'GET', {})
    v1_public_get_instruments_instrument = v1PublicGetInstrumentsInstrument = Entry('instruments/{instrument}', ['v1', 'public'], 'GET', {})
    v1_public_get_instruments_instrument_quote = v1PublicGetInstrumentsInstrumentQuote = Entry('instruments/{instrument}/quote', ['v1', 'public'], 'GET', {})
    v1_public_get_instruments_instrument_funding = v1PublicGetInstrumentsInstrumentFunding = Entry('instruments/{instrument}/funding', ['v1', 'public'], 'GET', {})
    v1_public_get = v1PublicGet = Entry('', ['v1', 'public'], 'GET', {})
    v1_private_get_orders = v1PrivateGetOrders = Entry('orders', ['v1', 'private'], 'GET', {})
    v1_private_get_orders_id = v1PrivateGetOrdersId = Entry('orders/{id}', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios = v1PrivateGetPortfolios = Entry('portfolios', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio = v1PrivateGetPortfoliosPortfolio = Entry('portfolios/{portfolio}', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio_detail = v1PrivateGetPortfoliosPortfolioDetail = Entry('portfolios/{portfolio}/detail', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio_summary = v1PrivateGetPortfoliosPortfolioSummary = Entry('portfolios/{portfolio}/summary', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio_balances = v1PrivateGetPortfoliosPortfolioBalances = Entry('portfolios/{portfolio}/balances', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio_balances_asset = v1PrivateGetPortfoliosPortfolioBalancesAsset = Entry('portfolios/{portfolio}/balances/{asset}', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio_positions = v1PrivateGetPortfoliosPortfolioPositions = Entry('portfolios/{portfolio}/positions', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio_positions_instrument = v1PrivateGetPortfoliosPortfolioPositionsInstrument = Entry('portfolios/{portfolio}/positions/{instrument}', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_fills = v1PrivateGetPortfoliosFills = Entry('portfolios/fills', ['v1', 'private'], 'GET', {})
    v1_private_get_portfolios_portfolio_fills = v1PrivateGetPortfoliosPortfolioFills = Entry('portfolios/{portfolio}/fills', ['v1', 'private'], 'GET', {})
    v1_private_get_transfers = v1PrivateGetTransfers = Entry('transfers', ['v1', 'private'], 'GET', {})
    v1_private_get_transfers_transfer_uuid = v1PrivateGetTransfersTransferUuid = Entry('transfers/{transfer_uuid}', ['v1', 'private'], 'GET', {})
    v1_private_post_orders = v1PrivatePostOrders = Entry('orders', ['v1', 'private'], 'POST', {})
    v1_private_post_portfolios = v1PrivatePostPortfolios = Entry('portfolios', ['v1', 'private'], 'POST', {})
    v1_private_post_portfolios_margin = v1PrivatePostPortfoliosMargin = Entry('portfolios/margin', ['v1', 'private'], 'POST', {})
    v1_private_post_portfolios_transfer = v1PrivatePostPortfoliosTransfer = Entry('portfolios/transfer', ['v1', 'private'], 'POST', {})
    v1_private_post_transfers_withdraw = v1PrivatePostTransfersWithdraw = Entry('transfers/withdraw', ['v1', 'private'], 'POST', {})
    v1_private_post_transfers_address = v1PrivatePostTransfersAddress = Entry('transfers/address', ['v1', 'private'], 'POST', {})
    v1_private_post_transfers_create_counterparty_id = v1PrivatePostTransfersCreateCounterpartyId = Entry('transfers/create-counterparty-id', ['v1', 'private'], 'POST', {})
    v1_private_post_transfers_validate_counterparty_id = v1PrivatePostTransfersValidateCounterpartyId = Entry('transfers/validate-counterparty-id', ['v1', 'private'], 'POST', {})
    v1_private_post_transfers_withdraw_counterparty = v1PrivatePostTransfersWithdrawCounterparty = Entry('transfers/withdraw/counterparty', ['v1', 'private'], 'POST', {})
    v1_private_put_orders_id = v1PrivatePutOrdersId = Entry('orders/{id}', ['v1', 'private'], 'PUT', {})
    v1_private_put_portfolios_portfolio = v1PrivatePutPortfoliosPortfolio = Entry('portfolios/{portfolio}', ['v1', 'private'], 'PUT', {})
    v1_private_delete_orders = v1PrivateDeleteOrders = Entry('orders', ['v1', 'private'], 'DELETE', {})
    v1_private_delete_orders_id = v1PrivateDeleteOrdersId = Entry('orders/{id}', ['v1', 'private'], 'DELETE', {})



"""A faster version of aiohttp's websocket client that uses select and other optimizations"""

import asyncio
import socket
import collections
from ccxt.async_support.base.ws.aiohttp_client import AiohttpClient


class FastClient(AiohttpClient):
    transport = None

    def __init__(self, url, on_message_callback, on_error_callback, on_close_callback, on_connected_callback, config={}):
        super(FastClient, self).__init__(url, on_message_callback, on_error_callback, on_close_callback, on_connected_callback, config)
        # instead of using the deque in aiohttp we implement our own for speed
        # https://github.com/aio-libs/aiohttp/blob/1d296d549050aa335ef542421b8b7dad788246d5/aiohttp/streams.py#L534
        self.stack = collections.deque()
        self.callback_scheduled = False

    def receive_loop(self):
        def handler():
            if not self.stack:
                self.callback_scheduled = False
                return
            message = self.stack.popleft()
            try:
                self.handle_message(message)
            except Exception as error:
                self.reject(error)
            self.asyncio_loop.call_soon(handler)

        def feed_data(message, size):
            if not self.callback_scheduled:
                self.callback_scheduled = True
                self.asyncio_loop.call_soon(handler)
            self.stack.append(message)

        def feed_eof():
            if self._close_code == 1000:  # OK close
                self.on_close(1000)
            else:
                self.on_error(1006)  # ABNORMAL_CLOSURE

        def wrapper(func):
            def parse_frame(buf):
                while self.stack:
                    self.handle_message(self.stack.popleft())
                return func(buf)
            return parse_frame

        async def close(code=1000, message=b''):
            # this is needed because our other wrappers break the closing process
            # we also don't wait for a response to the close message to speed it up
            # this code is adapted from aiohttp client_ws.py
            _self = self.connection
            if not _self._closed:
                _self._cancel_heartbeat()
                _self._closed = True
                try:
                    await _self._writer.close(code, message)
                    _self._response.close()
                    self._close_code = 1000
                except asyncio.CancelledError:
                    _self._response.close()
                    _self._close_code = 1006
                    raise
                except Exception as exc:
                    _self._close_code = 1006
                    _self._exception = exc
            return True

        connection = self.connection._conn
        if connection.closed:
            # connection got terminated after the connection was made and before the receive loop ran
            self.on_close(1006)
            return
        self.transport = connection.transport
        # increase the RCVBUF so that the tcp window size can be larger
        sock = self.transport.get_extra_info('socket')
        current_size = sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
        # 2 mebibytes is a performance value
        new_size = max(current_size, 2097152)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, new_size)

        ws_reader = connection.protocol._payload_parser
        ws_reader.parse_frame = wrapper(ws_reader.parse_frame)
        ws_reader.queue.feed_data = feed_data
        ws_reader.queue.feed_eof = feed_eof
        self.connection.close = close
        # return a future so super class won't complain
        return asyncio.sleep(0)

    def reset(self, error):
        super(FastClient, self).reset(error)
        self.stack.clear()
        if self.transport:
            self.transport.abort()



import os
import sys

root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
sys.path.append(root)

# ----------------------------------------------------------------------------

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

# ----------------------------------------------------------------------------

import ccxt  # noqa: F402
import hashlib  # noqa: F402

Exchange = ccxt.Exchange
hash = Exchange.hash
hmac = Exchange.hmac
ecdsa = Exchange.ecdsa
eddsa = Exchange.eddsa
jwt = Exchange.jwt
crc32 = Exchange.crc32
rsa = Exchange.rsa
encode = Exchange.encode


def equals(a, b):
    return a == b


exchange = Exchange()

# ---------------------------------------------------------------------------------------------------------------------

assert hash(encode(''), 'sha256', 'hex') == 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
assert hash(encode('cheese'), 'sha256', 'hex') == '873ac9ffea4dd04fa719e8920cd6938f0c23cd678af330939cff53c3d2855f34'

assert hash(encode(''), 'md5', 'hex') == 'd41d8cd98f00b204e9800998ecf8427e'
assert hash(encode('sexyfish'), 'md5', 'hex') == 'c8a35464aa9d5683585786f44d5889f8'

assert hash(encode(''), 'sha1', 'hex') == 'da39a3ee5e6b4b0d3255bfef95601890afd80709'
assert hash(encode('nutella'), 'sha1', 'hex') == 'b3d60a34b744159793c483b067c56d8affc5111a'
assert hmac(encode('hello'), encode('there'), hashlib.sha256, 'hex') == '551e1c1ecbce0fe9b643745a376584a6289f5f43a46861b315fac9edc8d52a26'
assert hmac(encode('a message'), encode('a secret'), hashlib.md5, 'hex') == '0bfa503bdbc7358185fcd49b4869e23d'

# ---------------------------------------------------------------------------------------------------------------------


privateKey = '1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a'


assert(equals(ecdsa('1a', privateKey, 'secp256k1', 'sha256'), {
    'r': '23dcb2a2a3728a35eb1a35cc01743c4609550d9cceaf2083550f13a9eb135f9f',
    's': '317963fcac18e4ec9f7921b97d7ea0c82a873dd6299cbfb6af016e08ef5ed667',
    'v': 0,
}))


assert(equals(ecdsa(privateKey, privateKey, 'secp256k1', None), {
    'r': 'b84a36a6fbabd5277ede578448b93d48e70b38efb5b15b1d4e2a298accf938b1',
    's': '66ebfb8221cda925526e699a59cd221bb4cc84bdc563024b1802c4d9e1d8bbe9',
    'v': 1,
}))

# ---------------------------------------------------------------------------------------------------------------------

#
# assert exchange.hashMessage(privateKey) == '0x59ea5d98c3500c3729f95cf98aa91663f498518cc401360df2912742c232207f'
#
# assert(equals(exchange.signHash('0x59ea5d98c3500c3729f95cf98aa91663f498518cc401360df2912742c232207f', privateKey), {
#     'r': '0x6f684aa41c02da83dac3039d8805ddbe79a03b1297e247c7742cab8dfc19d341',
#     's': '0x62473881674550563cb028ff40a7846fd53620ddf40a20cc1003b8484a109a4a',
#     'v': 27
# }))
#
# assert(equals(exchange.signMessage(privateKey, privateKey), {
#     'r': '0x6f684aa41c02da83dac3039d8805ddbe79a03b1297e247c7742cab8dfc19d341',
#     's': '0x62473881674550563cb028ff40a7846fd53620ddf40a20cc1003b8484a109a4a',
#     'v': 27
# }))
#
# ---------------------------------------------------------------------------------------------------------------------

pemKeyArray = [
    '-----BEGIN RSA PRIVATE KEY-----',
    'MIIEpAIBAAKCAQEAqUtXQXv2uSm9zPvdJTRpu5/65rBjAoHmIiowAs+u7fY0QP9O',
    '+T8CZRQjvZrfPomBiccjCxDo8mm7GkmL67rs9s7VOMQFvpTTtMBeglNSEbXJ3wTt',
    '7WM5gD3ZAbhfGtwWdGpcOMX0hV5/d/fiw1zQk0+AooayTryZ7HV+DQcD1sGTYYir',
    'LePUGkKV2NCvKtGOQfRAtuo0ccigMxMvEhWuExCc9Fhd077OpXMc0n7OiFqA2JWN',
    'jlocM9S71Srvw/lDos8a7lGaFUV8gP1CsQY/4qTju6hhsRd4lFujVr5J9FqmDbnv',
    'wQ79SSu6+lT0+ToFdDHiYOorBK2ESFR7EzlyLwIDAQABAoIBAEobQMbZjNbg/sSM',
    'O/HdT6tiDGKPM8gVNLgf34RbhSeFbrpFCDzy6Al3F24YLUEi0CGPmjdt34q93blU',
    'GHvIB5LCV3PR2vHiFAo7ayOBdZtrCEMn1T7lAHaynBu0qW0IiovLQzNW9AKtqv7I',
    '8+qw5lyVoKmEbOkqhfaMN/Fb8MJAo3yEVyhyp4EjtFhqDQ9DuHEjp1jGthnMgm+4',
    'qFyELf0DXZSAN+R9IGDqWPyo78lOcPW4pgeLuoLQ6Nn/2JAEGHFt1a7zMEw5yYbR',
    'XB89Bh0dheSmopBN6Heay4YuhnNKua37OlL1/nhKpLGUNE3z/UOXeiUiMeE1C2i+',
    'GjK+xgECgYEA4XH63rJ3VYDHHWUGlWffFWkEpgR4BDEkZ8MyVYQoUjQA2TpCvQ2+',
    'bR7aAIGXhBnRTrIBNu9m4+am0aUfi1hVmVHqt1b2vFgEb9uKuO4U62tCFpeji+Rz',
    '8C8QJyu5v5OUWdQZ4EA0d7ljeoTl50g8tpcGsDhnImLt+/jvJuFt8dECgYEAwD0n',
    'rhlkEjKVHxu1xwKeczxYCqqcUV9Quq8HnPh4VJQ77ljfZ57V7sAKrFDi5XfVD9jS',
    'oJPwxjw6za134VobhdvCM6WNs6AJmbR8E/b5kDMbGaZn27vt+/8JfnGaWBqpDi6W',
    'a5xoJOqmcMBQBK1YleZDoY4PLbk+onVKLMfGI/8CgYEAkyUIv8euGdGWpHnm5SdJ',
    'tLi5vv4Vs267uzntJWG/y3+DukTLgIdy7dgAI+pxkVgkg/+syUVSW5eU9CqZPyLl',
    'o8+Sqh2Jp36vTq71iSRj5RA5r3ND3K+8eFzPZzGj6AWUA1lrljFxzV7kLfiF8gH1',
    'FpvWUrhNoGT/vcFJno/uabECgYAecRC5hxfLserfVDoC261PvjyK492BHUDhbxob',
    'h1U2v4qGAdjOxd5Gwm5uPxjPEZzRt5oTB5pXKe5953xWWTiGh/hGyW6ZBTy/9E65',
    'sqBub0lZVHqZ1zamcwqD1WWFkiM3NbVoMQpk3iuhKzMAqpqekiofiSlqKi16+GvY',
    'j4IW7QKBgQCrIUckPZ6IY3ERIN1IL4TIcK2gOJcznp7fLWpC+sv35ya3OhtDXKFt',
    'GjIHmwLuiUNc0iPzA9Rw89W0zIkKWmWcxM88/ithdxh3MEeNDUGSnd4hQ9SECwxB',
    'Wem3eBT7I4VtFYoaTE3/bX1SKfgBdTzIRqWKSDpgBNZg/P2Tc+s11g==',
    '-----END RSA PRIVATE KEY-----',
]


pemKey = "\n".join(pemKeyArray)
assert rsa('hello', pemKey, 'sha256') == 'PqHotvSEBvM/AejnMOWBXUcOf3uHtcGu2zAYdlYdnlNSSQ80Uq4lcyAEstnZ2AnQJ9l5TCC53uoRZ26GQ47zlACgqtYglmPhQKLvQ5fldRzeBauYhGgM2C0mUuUGxh074fNGbK+bgmwEmDMIrnSPtXwiCqTAHh+8VEnC7us3t09D61y298dPBJYEBNN3dFZT0w0pCIQg3j3DSiFJOCfywmOKyXqS1pvmk6A38DVclQZORQ5WZXp2yvSKRLjxpzjxDl76h1GfbBl7sMLEFMyzk0wyIhIz8ZELMibYn036G4X1IcSlDcimthEkIbn2QjM0ntyYDZIS4QnsMBjvkV2UHw=='
assert rsa('poopy', pemKey, 'sha256') == 'or069qHwRDyl162T1s5G5+LfLnvDxlgk9kEsJvwI3vM02KB1LHW4+8gWqsV4TENZpeqed2Tb4na6ex+L/UR8JxJnnZtpVp33nUBgcp3miqp/YhcGN4++qolP4YN/21/AyfLFZW+VYggc+Mhh6PJSm+0dSEWMVsP35uH+35abXVxgB8GVOn7YTOSPaL8aw7hn4wlZWf2ieikKAi7AwAjkxnd7/Bu5+cW8D+ZdPHQfSKj+XHuPXlJNuQX0MDIqhdD2yuYJOQL56eKYs6nHPlClATkndSaAemQSGKet3X3Iz1awG4MGgz2Ei6bOanlNugc0f6Rng6rdwmqiMU3G4it6cw=='

assert jwt({'chicken': 'salad'}, encode(pemKey), 'sha256', True) == 'eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJjaGlja2VuIjoic2FsYWQifQ.FSKD5Y6RNzkHTuHdvG3753U7QNZ-u-GUSPfP1FMjEaK0Rr_iyQTSSmHhkdYSFFnmBvrrN_l-UwKwir52WlsgmQm9HYm0kidxbj7fWwrK2E1oe0P7OjupFjv1BZxc5W69WeaHtOPWe28tiHiON1LCnax6HgfI5lcIBsESGIIBZMVeaioQn9gDVwea7JxJvAlrhDIWZowIHTIdCQocXip7g5jREWHeEIuJNug67mwnfAFxCjvTRiTd0Bw6oBwjM3FLya-RyEyWrejQOWSuC8CNWVUHISaSmEyZ7uM6wTi2m_58TaE9mQwlef32OPErPvvBpgL5pZIyQ4ymwrCIFQLBQQ'
assert jwt({'lil': 'xan'}, encode('betrayed'), 'sha256', False) == 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJsaWwiOiJ4YW4ifQ.md-oFvZagA-NXmZoRNyJOQ7zwK-PWUMmMQ_LI9ZOKaM'

assert crc32('hello', True) == 907060870
assert crc32('tasty chicken breast :)', True) == 825820175
assert crc32('21101:0.00123125:21102:-0.001:21100:0.710705:21103:-0.001:21096:0.71076:21104:-0.001:21094:1.0746:21105:-0.001:21093:0.710854:21106:-0.419:21092:0.01368102:21107:-0.001:21090:0.710975:21109:-0.001:21089:0.63586344:21110:-1.186213:21087:0.299:21111:-0.48751202:21086:0.9493:21112:-0.03702409:21082:0.03537667:21113:-0.712385:21081:0.00101366:21114:-0.2903:21079:0.710713:21115:-0.001:21078:0.997048:21116:-0.60089827:21077:0.23770225:21117:-0.83201:21076:0.03619135:21118:-0.09996142:21075:0.1272433:21119:-1.09681107:21074:0.7447885:21120:-0.04771792:21073:0.0011:21121:-0.91495684:21072:0.73311632:21122:-0.07940416:21071:0.09817:21123:-0.39376843:21070:0.19101052:21124:-1.51692599:21069:0.2757:21125:-0.11107322:21068:0.12480303:21126:-0.12704666:21067:0.4201:21128:-0.12804666', True) == -51055998

# assert eddsa('1b1b', privateKey, 'ed25519') == '3DBaaz8z4Pq9n6ncNCjB4pFLWaWTXbjaCUqKQmBgS3w7AP6opeDqANBhPssbV3jyfJB4LfK8kGR6pu6GU8fbjMuy'



from ccxt.base.types import Entry


class ImplicitAPI:
    public_post_info = publicPostInfo = Entry('info', 'public', 'POST', {'cost': 1})
    private_post_exchange = privatePostExchange = Entry('exchange', 'private', 'POST', {'cost': 1})



# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

import ccxt.async_support
from ccxt.async_support.base.ws.cache import ArrayCache, ArrayCacheBySymbolById, ArrayCacheByTimestamp
import hashlib
from ccxt.base.types import Balances, Int, Order, OrderBook, Str, Ticker, Trade
from ccxt.async_support.base.ws.client import Client
from typing import List
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import ArgumentsRequired
from ccxt.base.errors import NotSupported


class deribit(ccxt.async_support.deribit):

    def describe(self):
        return self.deep_extend(super(deribit, self).describe(), {
            'has': {
                'ws': True,
                'watchBalance': True,
                'watchTicker': True,
                'watchTickers': False,
                'watchTrades': True,
                'watchTradesForSymbols': True,
                'watchMyTrades': True,
                'watchOrders': True,
                'watchOrderBook': True,
                'watchOrderBookForSymbols': True,
                'watchOHLCV': True,
                'watchOHLCVForSymbols': True,
            },
            'urls': {
                'test': {
                    'ws': 'wss://test.deribit.com/ws/api/v2',
                },
                'api': {
                    'ws': 'wss://www.deribit.com/ws/api/v2',
                },
            },
            'options': {
                'ws': {
                    'timeframes': {
                        '1m': '1',
                        '3m': '3',
                        '5m': '5',
                        '15m': '15',
                        '30m': '30',
                        '1h': '60',
                        '2h': '120',
                        '4h': '180',
                        '6h': '360',
                        '12h': '720',
                        '1d': '1D',
                    },
                    # watchTrades replacement
                    'watchTradesForSymbols': {
                        'interval': '100ms',  # 100ms, agg2, raw
                    },
                    # watchOrderBook replacement
                    'watchOrderBookForSymbols': {
                        'interval': '100ms',  # 100ms, agg2, raw
                        'useDepthEndpoint': False,  # if True, it will use the {books.group.depth.interval} endpoint instead of the {books.interval} endpoint
                        'depth': '20',  # 1, 10, 20
                        'group': 'none',  # none, 1, 2, 5, 10, 25, 100, 250
                    },
                },
                'currencies': ['BTC', 'ETH', 'SOL', 'USDC'],
            },
            'streaming': {
            },
            'exceptions': {
            },
        })

    def request_id(self):
        requestId = self.sum(self.safe_integer(self.options, 'requestId', 0), 1)
        self.options['requestId'] = requestId
        return requestId

    async def watch_balance(self, params={}) -> Balances:
        """
        :see: https://docs.deribit.com/#user-portfolio-currency
        watch balance and get the amount of funds available for trading or funds locked in orders
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `balance structure <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        await self.authenticate(params)
        messageHash = 'balance'
        url = self.urls['api']['ws']
        currencies = self.safe_value(self.options, 'currencies', [])
        channels = []
        for i in range(0, len(currencies)):
            currencyCode = currencies[i]
            channels.append('user.portfolio.' + currencyCode)
        subscribe: dict = {
            'jsonrpc': '2.0',
            'method': 'private/subscribe',
            'params': {
                'channels': channels,
            },
            'id': self.request_id(),
        }
        request = self.deep_extend(subscribe, params)
        return await self.watch(url, messageHash, request, messageHash, request)

    def handle_balance(self, client: Client, message):
        #
        # subscription
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "user.portfolio.btc",
        #             "data": {
        #                 "total_pl": 0,
        #                 "session_upl": 0,
        #                 "session_rpl": 0,
        #                 "projected_maintenance_margin": 0,
        #                 "projected_initial_margin": 0,
        #                 "projected_delta_total": 0,
        #                 "portfolio_margining_enabled": False,
        #                 "options_vega": 0,
        #                 "options_value": 0,
        #                 "options_theta": 0,
        #                 "options_session_upl": 0,
        #                 "options_session_rpl": 0,
        #                 "options_pl": 0,
        #                 "options_gamma": 0,
        #                 "options_delta": 0,
        #                 "margin_balance": 0.0015,
        #                 "maintenance_margin": 0,
        #                 "initial_margin": 0,
        #                 "futures_session_upl": 0,
        #                 "futures_session_rpl": 0,
        #                 "futures_pl": 0,
        #                 "fee_balance": 0,
        #                 "estimated_liquidation_ratio_map": {},
        #                 "estimated_liquidation_ratio": 0,
        #                 "equity": 0.0015,
        #                 "delta_total_map": {},
        #                 "delta_total": 0,
        #                 "currency": "BTC",
        #                 "balance": 0.0015,
        #                 "available_withdrawal_funds": 0.0015,
        #                 "available_funds": 0.0015
        #             }
        #         }
        #     }
        #
        params = self.safe_value(message, 'params', {})
        data = self.safe_value(params, 'data', {})
        self.balance['info'] = data
        currencyId = self.safe_string(data, 'currency')
        currencyCode = self.safe_currency_code(currencyId)
        balance = self.parse_balance(data)
        self.balance[currencyCode] = balance
        messageHash = 'balance'
        client.resolve(self.balance, messageHash)

    async def watch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        :see: https://docs.deribit.com/#ticker-instrument_name-interval
        watches a price ticker, a statistical calculation with the information for a specific market.
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.interval]: specify aggregation and frequency of notifications. Possible values: 100ms, raw
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        market = self.market(symbol)
        url = self.urls['api']['ws']
        interval = self.safe_string(params, 'interval', '100ms')
        params = self.omit(params, 'interval')
        await self.load_markets()
        if interval == 'raw':
            await self.authenticate()
        channel = 'ticker.' + market['id'] + '.' + interval
        message: dict = {
            'jsonrpc': '2.0',
            'method': 'public/subscribe',
            'params': {
                'channels': ['ticker.' + market['id'] + '.' + interval],
            },
            'id': self.request_id(),
        }
        request = self.deep_extend(message, params)
        return await self.watch(url, channel, request, channel, request)

    def handle_ticker(self, client: Client, message):
        #
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "ticker.BTC_USDC-PERPETUAL.raw",
        #             "data": {
        #                 "timestamp": 1655393725040,
        #                 "stats": [Object],
        #                 "state": "open",
        #                 "settlement_price": 21729.5891,
        #                 "open_interest": 164.501,
        #                 "min_price": 20792.9376,
        #                 "max_price": 21426.225,
        #                 "mark_price": 21109.555,
        #                 "last_price": 21132,
        #                 "instrument_name": "BTC_USDC-PERPETUAL",
        #                 "index_price": 21122.3937,
        #                 "funding_8h": -0.00022427,
        #                 "estimated_delivery_price": 21122.3937,
        #                 "current_funding": -0.00010782,
        #                 "best_bid_price": 21106,
        #                 "best_bid_amount": 1.143,
        #                 "best_ask_price": 21113,
        #                 "best_ask_amount": 0.327
        #             }
        #         }
        #     }
        #
        params = self.safe_value(message, 'params', {})
        data = self.safe_value(params, 'data', {})
        marketId = self.safe_string(data, 'instrument_name')
        symbol = self.safe_symbol(marketId)
        ticker = self.parse_ticker(data)
        messageHash = self.safe_string(params, 'channel')
        self.tickers[symbol] = ticker
        client.resolve(ticker, messageHash)

    async def watch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a particular symbol
        :see: https://docs.deribit.com/#trades-instrument_name-interval
        :param str symbol: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.interval]: specify aggregation and frequency of notifications. Possible values: 100ms, raw
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        params['callerMethodName'] = 'watchTrades'
        return await self.watch_trades_for_symbols([symbol], since, limit, params)

    async def watch_trades_for_symbols(self, symbols: List[str], since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a list of symbols
        :see: https://docs.deribit.com/#trades-instrument_name-interval
        :param str[] symbols: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        interval = None
        interval, params = self.handle_option_and_params(params, 'watchTradesForSymbols', 'interval', '100ms')
        if interval == 'raw':
            await self.authenticate()
        trades = await self.watch_multiple_wrapper('trades', interval, symbols, params)
        if self.newUpdates:
            first = self.safe_dict(trades, 0)
            tradeSymbol = self.safe_string(first, 'symbol')
            limit = trades.getLimit(tradeSymbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp', True)

    def handle_trades(self, client: Client, message):
        #
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "trades.BTC_USDC-PERPETUAL.100ms",
        #             "data": [{
        #                 "trade_seq": 501899,
        #                 "trade_id": "USDC-2436803",
        #                 "timestamp": 1655397355998,
        #                 "tick_direction": 2,
        #                 "price": 21026,
        #                 "mark_price": 21019.9719,
        #                 "instrument_name": "BTC_USDC-PERPETUAL",
        #                 "index_price": 21031.7847,
        #                 "direction": "buy",
        #                 "amount": 0.049
        #             }]
        #         }
        #     }
        #
        params = self.safe_dict(message, 'params', {})
        channel = self.safe_string(params, 'channel', '')
        parts = channel.split('.')
        marketId = self.safe_string(parts, 1)
        interval = self.safe_string(parts, 2)
        symbol = self.safe_symbol(marketId)
        market = self.safe_market(marketId)
        trades = self.safe_list(params, 'data', [])
        if self.safe_value(self.trades, symbol) is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            self.trades[symbol] = ArrayCache(limit)
        stored = self.trades[symbol]
        for i in range(0, len(trades)):
            trade = trades[i]
            parsed = self.parse_trade(trade, market)
            stored.append(parsed)
        self.trades[symbol] = stored
        messageHash = 'trades|' + symbol + '|' + interval
        client.resolve(self.trades[symbol], messageHash)

    async def watch_my_trades(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of trades associated with the user
        :see: https://docs.deribit.com/#user-trades-instrument_name-interval
        :param str symbol: unified symbol of the market to fetch trades for. Use 'any' to watch all trades
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.interval]: specify aggregation and frequency of notifications. Possible values: 100ms, raw
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        await self.authenticate(params)
        if symbol is not None:
            await self.load_markets()
            symbol = self.symbol(symbol)
        url = self.urls['api']['ws']
        interval = self.safe_string(params, 'interval', 'raw')
        params = self.omit(params, 'interval')
        channel = 'user.trades.any.any.' + interval
        message: dict = {
            'jsonrpc': '2.0',
            'method': 'private/subscribe',
            'params': {
                'channels': [channel],
            },
            'id': self.request_id(),
        }
        request = self.deep_extend(message, params)
        trades = await self.watch(url, channel, request, channel, request)
        return self.filter_by_symbol_since_limit(trades, symbol, since, limit, True)

    def handle_my_trades(self, client: Client, message):
        #
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "user.trades.any.any.raw",
        #             "data": [{
        #                 "trade_seq": 149546319,
        #                 "trade_id": "219381310",
        #                 "timestamp": 1655421193564,
        #                 "tick_direction": 0,
        #                 "state": "filled",
        #                 "self_trade": False,
        #                 "reduce_only": False,
        #                 "profit_loss": 0,
        #                 "price": 20236.5,
        #                 "post_only": False,
        #                 "order_type": "market",
        #                 "order_id": "46108941243",
        #                 "matching_id": null,
        #                 "mark_price": 20233.96,
        #                 "liquidity": "T",
        #                 "instrument_name": "BTC-PERPETUAL",
        #                 "index_price": 20253.31,
        #                 "fee_currency": "BTC",
        #                 "fee": 2.5e-7,
        #                 "direction": "buy",
        #                 "amount": 10
        #             }]
        #         }
        #     }
        #
        params = self.safe_value(message, 'params', {})
        channel = self.safe_string(params, 'channel', '')
        trades = self.safe_value(params, 'data', [])
        cachedTrades = self.myTrades
        if cachedTrades is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            cachedTrades = ArrayCacheBySymbolById(limit)
        parsed = self.parse_trades(trades)
        marketIds: dict = {}
        for i in range(0, len(parsed)):
            trade = parsed[i]
            cachedTrades.append(trade)
            symbol = trade['symbol']
            marketIds[symbol] = True
        client.resolve(cachedTrades, channel)

    async def watch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        :see: https://docs.deribit.com/#book-instrument_name-group-depth-interval
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.interval]: Frequency of notifications. Events will be aggregated over self interval. Possible values: 100ms, raw
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        params['callerMethodName'] = 'watchOrderBook'
        return await self.watch_order_book_for_symbols([symbol], limit, params)

    async def watch_order_book_for_symbols(self, symbols: List[str], limit: Int = None, params={}) -> OrderBook:
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :see: https://docs.deribit.com/#book-instrument_name-group-depth-interval
        :param str[] symbols: unified array of symbols
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        interval = None
        interval, params = self.handle_option_and_params(params, 'watchOrderBookForSymbols', 'interval', '100ms')
        if interval == 'raw':
            await self.authenticate()
        descriptor = ''
        useDepthEndpoint = None  # for more info, see comment in .options
        useDepthEndpoint, params = self.handle_option_and_params(params, 'watchOrderBookForSymbols', 'useDepthEndpoint', False)
        if useDepthEndpoint:
            depth = None
            depth, params = self.handle_option_and_params(params, 'watchOrderBookForSymbols', 'depth', '20')
            group = None
            group, params = self.handle_option_and_params(params, 'watchOrderBookForSymbols', 'group', 'none')
            descriptor = group + '.' + depth + '.' + interval
        else:
            descriptor = interval
        orderbook = await self.watch_multiple_wrapper('book', descriptor, symbols, params)
        return orderbook.limit()

    def handle_order_book(self, client: Client, message):
        #
        #  snapshot
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "book.BTC_USDC-PERPETUAL.raw",
        #             "data": {
        #                 "type": "snapshot",
        #                 "timestamp": 1655395057025,
        #                 "instrument_name": "BTC_USDC-PERPETUAL",
        #                 "change_id": 1550694837,
        #                 "bids": [
        #                     ["new", 20987, 0.487],
        #                     ["new", 20986, 0.238],
        #                 ],
        #                 "asks": [
        #                     ["new", 20999, 0.092],
        #                     ["new", 21000, 1.238],
        #                 ]
        #             }
        #         }
        #     }
        #
        #  change
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "book.BTC_USDC-PERPETUAL.raw",
        #             "data": {
        #                 "type": "change",
        #                 "timestamp": 1655395168086,
        #                 "prev_change_id": 1550724481,
        #                 "instrument_name": "BTC_USDC-PERPETUAL",
        #                 "change_id": 1550724483,
        #                 "bids": [
        #                     ["new", 20977, 0.109],
        #                     ["delete", 20975, 0]
        #                 ],
        #                 "asks": []
        #             }
        #         }
        #     }
        #
        params = self.safe_value(message, 'params', {})
        data = self.safe_value(params, 'data', {})
        channel = self.safe_string(params, 'channel')
        parts = channel.split('.')
        descriptor = ''
        partsLength = len(parts)
        isDetailed = partsLength == 5
        if isDetailed:
            group = self.safe_string(parts, 2)
            depth = self.safe_string(parts, 3)
            interval = self.safe_string(parts, 4)
            descriptor = group + '.' + depth + '.' + interval
        else:
            interval = self.safe_string(parts, 2)
            descriptor = interval
        marketId = self.safe_string(data, 'instrument_name')
        symbol = self.safe_symbol(marketId)
        timestamp = self.safe_integer(data, 'timestamp')
        if not (symbol in self.orderbooks):
            self.orderbooks[symbol] = self.counted_order_book()
        storedOrderBook = self.orderbooks[symbol]
        asks = self.safe_list(data, 'asks', [])
        bids = self.safe_list(data, 'bids', [])
        self.handle_deltas(storedOrderBook['asks'], asks)
        self.handle_deltas(storedOrderBook['bids'], bids)
        storedOrderBook['nonce'] = timestamp
        storedOrderBook['timestamp'] = timestamp
        storedOrderBook['datetime'] = self.iso8601(timestamp)
        storedOrderBook['symbol'] = symbol
        self.orderbooks[symbol] = storedOrderBook
        messageHash = 'book|' + symbol + '|' + descriptor
        client.resolve(storedOrderBook, messageHash)

    def clean_order_book(self, data):
        bids = self.safe_list(data, 'bids', [])
        asks = self.safe_list(data, 'asks', [])
        cleanedBids = []
        for i in range(0, len(bids)):
            cleanedBids.append([bids[i][1], bids[i][2]])
        cleanedAsks = []
        for i in range(0, len(asks)):
            cleanedAsks.append([asks[i][1], asks[i][2]])
        data['bids'] = cleanedBids
        data['asks'] = cleanedAsks
        return data

    def handle_delta(self, bookside, delta):
        price = delta[1]
        amount = delta[2]
        if delta[0] == 'new' or delta[0] == 'change':
            bookside.storeArray([price, amount, 1])
        elif delta[0] == 'delete':
            bookside.storeArray([price, amount, 0])

    def handle_deltas(self, bookside, deltas):
        for i in range(0, len(deltas)):
            self.handle_delta(bookside, deltas[i])

    async def watch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        :see: https://docs.deribit.com/#user-orders-instrument_name-raw
        watches information on multiple orders made by the user
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of [order structures]{@link https://docs.ccxt.com/#/?id=order-structure
        """
        await self.load_markets()
        await self.authenticate(params)
        if symbol is not None:
            symbol = self.symbol(symbol)
        url = self.urls['api']['ws']
        currency = self.safe_string(params, 'currency', 'any')
        interval = self.safe_string(params, 'interval', 'raw')
        kind = self.safe_string(params, 'kind', 'any')
        params = self.omit(params, 'interval', 'currency', 'kind')
        channel = 'user.orders.' + kind + '.' + currency + '.' + interval
        message: dict = {
            'jsonrpc': '2.0',
            'method': 'private/subscribe',
            'params': {
                'channels': [channel],
            },
            'id': self.request_id(),
        }
        request = self.deep_extend(message, params)
        orders = await self.watch(url, channel, request, channel, request)
        if self.newUpdates:
            limit = orders.getLimit(symbol, limit)
        return self.filter_by_symbol_since_limit(orders, symbol, since, limit, True)

    def handle_orders(self, client: Client, message):
        # Does not return a snapshot of current orders
        #
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "user.orders.any.any.raw",
        #             "data": {
        #                 "web": True,
        #                 "time_in_force": "good_til_cancelled",
        #                 "replaced": False,
        #                 "reduce_only": False,
        #                 "profit_loss": 0,
        #                 "price": 50000,
        #                 "post_only": False,
        #                 "order_type": "limit",
        #                 "order_state": "open",
        #                 "order_id": "46094375191",
        #                 "max_show": 10,
        #                 "last_update_timestamp": 1655401625037,
        #                 "label": '',
        #                 "is_liquidation": False,
        #                 "instrument_name": "BTC-PERPETUAL",
        #                 "filled_amount": 0,
        #                 "direction": "sell",
        #                 "creation_timestamp": 1655401625037,
        #                 "commission": 0,
        #                 "average_price": 0,
        #                 "api": False,
        #                 "amount": 10
        #             }
        #         }
        #     }
        #
        if self.orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit', 1000)
            self.orders = ArrayCacheBySymbolById(limit)
        params = self.safe_value(message, 'params', {})
        channel = self.safe_string(params, 'channel', '')
        data = self.safe_value(params, 'data', {})
        orders = []
        if isinstance(data, list):
            orders = self.parse_orders(data)
        else:
            order = self.parse_order(data)
            orders = [order]
        cachedOrders = self.orders
        for i in range(0, len(orders)):
            cachedOrders.append(orders[i])
        client.resolve(self.orders, channel)

    async def watch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        :see: https://docs.deribit.com/#chart-trades-instrument_name-resolution
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        await self.load_markets()
        symbol = self.symbol(symbol)
        ohlcvs = await self.watch_ohlcv_for_symbols([[symbol, timeframe]], since, limit, params)
        return ohlcvs[symbol][timeframe]

    async def watch_ohlcv_for_symbols(self, symbolsAndTimeframes: List[List[str]], since: Int = None, limit: Int = None, params={}):
        """
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :see: https://docs.deribit.com/#chart-trades-instrument_name-resolution
        :param str[][] symbolsAndTimeframes: array of arrays containing unified symbols and timeframes to fetch OHLCV data for, example [['BTC/USDT', '1m'], ['LTC/USDT', '5m']]
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        symbolsLength = len(symbolsAndTimeframes)
        if symbolsLength == 0 or not isinstance(symbolsAndTimeframes[0], list):
            raise ArgumentsRequired(self.id + " watchOHLCVForSymbols() requires a an array of symbols and timeframes, like  [['BTC/USDT', '1m'], ['LTC/USDT', '5m']]")
        symbol, timeframe, candles = await self.watch_multiple_wrapper('chart.trades', None, symbolsAndTimeframes, params)
        if self.newUpdates:
            limit = candles.getLimit(symbol, limit)
        filtered = self.filter_by_since_limit(candles, since, limit, 0, True)
        return self.create_ohlcv_object(symbol, timeframe, filtered)

    def handle_ohlcv(self, client: Client, message):
        #
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "chart.trades.BTC_USDC-PERPETUAL.1",
        #             "data": {
        #                 "volume": 0,
        #                 "tick": 1655403420000,
        #                 "open": 20951,
        #                 "low": 20951,
        #                 "high": 20951,
        #                 "cost": 0,
        #                 "close": 20951
        #             }
        #         }
        #     }
        #
        params = self.safe_dict(message, 'params', {})
        channel = self.safe_string(params, 'channel', '')
        parts = channel.split('.')
        marketId = self.safe_string(parts, 2)
        rawTimeframe = self.safe_string(parts, 3)
        market = self.safe_market(marketId)
        symbol = market['symbol']
        wsOptions = self.safe_dict(self.options, 'ws', {})
        timeframes = self.safe_dict(wsOptions, 'timeframes', {})
        unifiedTimeframe = self.find_timeframe(rawTimeframe, timeframes)
        self.ohlcvs[symbol] = self.safe_dict(self.ohlcvs, symbol, {})
        if self.safe_value(self.ohlcvs[symbol], unifiedTimeframe) is None:
            limit = self.safe_integer(self.options, 'OHLCVLimit', 1000)
            self.ohlcvs[symbol][unifiedTimeframe] = ArrayCacheByTimestamp(limit)
        stored = self.ohlcvs[symbol][unifiedTimeframe]
        ohlcv = self.safe_dict(params, 'data', {})
        # data contains a single OHLCV candle
        parsed = self.parse_ws_ohlcv(ohlcv, market)
        stored.append(parsed)
        self.ohlcvs[symbol][unifiedTimeframe] = stored
        resolveData = [symbol, unifiedTimeframe, stored]
        messageHash = 'chart.trades|' + symbol + '|' + rawTimeframe
        client.resolve(resolveData, messageHash)

    def parse_ws_ohlcv(self, ohlcv, market=None) -> list:
        #
        #    {
        #        "c": "28909.0",
        #        "o": "28915.4",
        #        "h": "28915.4",
        #        "l": "28896.1",
        #        "v": "27.6919",
        #        "T": 1696687499999,
        #        "t": 1696687440000
        #    }
        #
        return [
            self.safe_integer(ohlcv, 'tick'),
            self.safe_number(ohlcv, 'open'),
            self.safe_number(ohlcv, 'high'),
            self.safe_number(ohlcv, 'low'),
            self.safe_number(ohlcv, 'close'),
            self.safe_number(ohlcv, 'volume'),
        ]

    async def watch_multiple_wrapper(self, channelName: str, channelDescriptor: Str, symbolsArray=None, params={}):
        await self.load_markets()
        url = self.urls['api']['ws']
        rawSubscriptions = []
        messageHashes = []
        isOHLCV = (channelName == 'chart.trades')
        symbols = self.get_list_from_object_values(symbolsArray, 0) if isOHLCV else symbolsArray
        self.market_symbols(symbols, None, False)
        for i in range(0, len(symbolsArray)):
            current = symbolsArray[i]
            market = None
            if isOHLCV:
                market = self.market(current[0])
                unifiedTf = current[1]
                rawTf = self.safe_string(self.timeframes, unifiedTf, unifiedTf)
                channelDescriptor = rawTf
            else:
                market = self.market(current)
            message = channelName + '.' + market['id'] + '.' + channelDescriptor
            rawSubscriptions.append(message)
            messageHashes.append(channelName + '|' + market['symbol'] + '|' + channelDescriptor)
        request: dict = {
            'jsonrpc': '2.0',
            'method': 'public/subscribe',
            'params': {
                'channels': rawSubscriptions,
            },
            'id': self.request_id(),
        }
        extendedRequest = self.deep_extend(request, params)
        maxMessageByteLimit = 32768 - 1  # 'Message Too Big: limit 32768B'
        jsonedText = self.json(extendedRequest)
        if len(jsonedText) >= maxMessageByteLimit:
            raise ExchangeError(self.id + ' requested subscription length over limit, try to reduce symbols amount')
        return await self.watch_multiple(url, messageHashes, extendedRequest, rawSubscriptions)

    def handle_message(self, client: Client, message):
        #
        # error
        #     {
        #         "jsonrpc": "2.0",
        #         "id": 1,
        #         "error": {
        #             "message": "Invalid params",
        #             "data": {
        #                 "reason": "invalid format",
        #                 "param": "nonce"
        #             },
        #             "code": -32602
        #         },
        #         "usIn": "1655391709417993",
        #         "usOut": "1655391709418049",
        #         "usDiff": 56,
        #         "testnet": False
        #     }
        #
        # subscribe
        #     {
        #         "jsonrpc": "2.0",
        #         "id": 2,
        #         "result": ["ticker.BTC_USDC-PERPETUAL.raw"],
        #         "usIn": "1655393625889396",
        #         "usOut": "1655393625889518",
        #         "usDiff": 122,
        #         "testnet": False
        #     }
        #
        # notification
        #     {
        #         "jsonrpc": "2.0",
        #         "method": "subscription",
        #         "params": {
        #             "channel": "ticker.BTC_USDC-PERPETUAL.raw",
        #             "data": {
        #                 "timestamp": 1655393724752,
        #                 "stats": [Object],
        #                 "state": "open",
        #                 "settlement_price": 21729.5891,
        #                 "open_interest": 164.501,
        #                 "min_price": 20792.9001,
        #                 "max_price": 21426.1864,
        #                 "mark_price": 21109.4757,
        #                 "last_price": 21132,
        #                 "instrument_name": "BTC_USDC-PERPETUAL",
        #                 "index_price": 21122.3937,
        #                 "funding_8h": -0.00022427,
        #                 "estimated_delivery_price": 21122.3937,
        #                 "current_funding": -0.00011158,
        #                 "best_bid_price": 21106,
        #                 "best_bid_amount": 1.143,
        #                 "best_ask_price": 21113,
        #                 "best_ask_amount": 0.402
        #             }
        #         }
        #     }
        #
        error = self.safe_value(message, 'error')
        if error is not None:
            raise ExchangeError(self.id + ' ' + self.json(error))
        params = self.safe_value(message, 'params')
        channel = self.safe_string(params, 'channel')
        if channel is not None:
            parts = channel.split('.')
            channelId = self.safe_string(parts, 0)
            userHandlers: dict = {
                'trades': self.handle_my_trades,
                'portfolio': self.handle_balance,
                'orders': self.handle_orders,
            }
            handlers: dict = {
                'ticker': self.handle_ticker,
                'book': self.handle_order_book,
                'trades': self.handle_trades,
                'chart': self.handle_ohlcv,
                'user': self.safe_value(userHandlers, self.safe_string(parts, 1)),
            }
            handler = self.safe_value(handlers, channelId)
            if handler is not None:
                handler(client, message)
                return
            raise NotSupported(self.id + ' no handler found for self message ' + self.json(message))
        result = self.safe_value(message, 'result', {})
        accessToken = self.safe_string(result, 'access_token')
        if accessToken is not None:
            self.handle_authentication_message(client, message)

    def handle_authentication_message(self, client: Client, message):
        #
        #     {
        #         "jsonrpc": "2.0",
        #         "id": 1,
        #         "result": {
        #             "token_type": "bearer",
        #             "scope": "account:read_write block_trade:read_write connection custody:read_write mainaccount name:ccxt trade:read_write wallet:read_write",
        #             "refresh_token": "1686927372328.1EzFBRmt.logRQWXkPA1oE_Tk0gRsls9Hau7YN6a321XUBnxvR4x6cryhbkKcniUJU-czA8_zKXrqQGpQmfoDwhLIjIsWCvRuu6otbg-LKWlrtTX1GQqLcPaTTHAdZGTMV-HM8HiS03QBd9MIXWRfF53sKj2hdR9nZPZ6MH1XrkpAZPB_peuEEB9wlcc3elzWEZFtCmiy1fnQ8TPHwAJMt3nuUmEcMLt_-F554qrsg_-I66D9xMiifJj4dBemdPfV_PkGPRIwIoKlxDjyv2-xfCw-4eKyo6Hu1m2h6gT1DPOTxSXcBgfBQjpi-_uY3iAIj7U6xjC46PHthEdquhEuCTZl7UfCRZSAWwZA",
        #             "expires_in": 31536000,
        #             "access_token": "1686923272328.1CkwEx-u.qHradpIulmuoeboKMEi8PkQ1_4DF8yFE2zywBTtkD32sruVC53b1HwL5OWRuh2nYAndXff4xuXIMRkkEfMAFCeq24prihxxinoS8DDVkKBxedGx4CUPJFeXjmh7wuRGqQOLg1plXOpbF3fwF2KPEkAuETwcpcVY6K9HUVjutNRfxFe2TR7CvuS9x8TATvoPeu7H1ezYl-LkKSaRifdTXuwituXgp4oDbPRyQLniEBWuYF9rY7qbABxuOJlXI1VZ63u7Bh0mGWei-KeVeqHGNpy6OgrFRPXPxa9_U7vaxCyHW3zZ9959TQ1QUMLWtUX-NLBEv3BT5eCieW9HORYIOKfsgkpd3"
        #         },
        #         "usIn": "1655391872327712",
        #         "usOut": "1655391872328515",
        #         "usDiff": 803,
        #         "testnet": False
        #     }
        #
        messageHash = 'authenticated'
        client.resolve(message, messageHash)
        return message

    async def authenticate(self, params={}):
        url = self.urls['api']['ws']
        client = self.client(url)
        time = self.milliseconds()
        timeString = self.number_to_string(time)
        nonce = timeString
        messageHash = 'authenticated'
        future = self.safe_value(client.subscriptions, messageHash)
        if future is None:
            self.check_required_credentials()
            requestId = self.request_id()
            signature = self.hmac(self.encode(timeString + '\n' + nonce + '\n'), self.encode(self.secret), hashlib.sha256)
            request: dict = {
                'jsonrpc': '2.0',
                'id': requestId,
                'method': 'public/auth',
                'params': {
                    'grant_type': 'client_signature',
                    'client_id': self.apiKey,
                    'timestamp': time,
                    'signature': signature,
                    'nonce': nonce,
                    'data': '',
                },
            }
            future = await self.watch(url, messageHash, self.extend(request, params), messageHash)
            client.subscriptions[messageHash] = future
        return future
